from __future__ import print_function

import logging
from copy import copy

import numpy as np

from ..IO.image_stack import ImageStack
from ..elements.b_splines import BSplineSurface


def make_grid(c1x, c1y, c2x, c2y, ny, nx, elm):
    """
    Makes regular grid for the given corner coordinates, number of elements along each axis and finite element
    definitions.




     Parameters
     ----------
    c1x : float
        X-position of upper left corner
    c1y : float
        Y-position of upper left corner
    c2x : float
        X-position of lower right corner
    c2y : float
        Y-position of lower right corner
    nx : int
        Number of elements along X-axis
    ny : int
        Number of elements along Y-axis
    elm : Element object
        The element definitions   

     Returns
     -------
    X-coordinates of nodes, Y-Coordinates of nodes
     """

    elm.set_n_nodes((nx, ny))

    elm_width = float(c2x - c1x)
    elm_heigt = float(c2y - c1y)

    # Scale one element to element width and height
    nodes_x = elm.ctrl_pos_e * elm_width
    nodes_y = elm.ctrl_pos_n * elm_heigt

    # Generate elements
    elements = list(zip((nodes_y), (nodes_x)))

    # Unpack nodes
    ynod, xnod = zip(*elements)

    # Shift nodes to global frame
    node_x = np.array(xnod) + c1x
    node_y = np.array(ynod) + c1y

    return node_x, node_y


class Mesher(object):
    def __init__(self, deg_e=1, deg_n=1):

        """
        Mesher utility

        The Mesher is used to generate an Mesh object and provides a lightweigt GUI.



         Parameters
         ----------
        deg_e : int
            The polynomial degree in the e-direction
        deg_n : int
            The polynomial degree in the n-direction

         Returns
         -------
        Mesh :  Mehser object
         """

        self.deg_e = deg_e
        self.deg_n = deg_n

    def __gui__(self):
        from matplotlib.widgets import Button, RectangleSelector
        import matplotlib.pyplot as plt
        plt.rcParams['font.size'] = 8

        def render_mesh():
            try:
                data.set_xdata(
                    self._mesh_.xnodes.transpose())
                data.set_ydata(
                    self._mesh_.ynodes.transpose())
                fig.canvas.draw()
            except:
                print('Could not render mesh')
                pass

        def line_select_callback(eclick, erelease):
            'eclick and erelease are the press and release events'
            x1, y1 = eclick.xdata, eclick.ydata
            x2, y2 = erelease.xdata, erelease.ydata

            self._mesh_.Xc1 = min([x1, x2])
            self._mesh_.Xc2 = max([x1, x2])
            self._mesh_.Yc1 = min([y1, y2])
            self._mesh_.Yc2 = max([y1, y2])

            self._mesh_.gen_node_positions()
            render_mesh()

        def toggle_selector(event):

            if event.key in ['W', 'w']:
                self._mesh_.n_ely += 1

            if event.key in ['X', 'x']:
                self._mesh_.n_ely -= 1

            if event.key in ['A', 'a']:
                self._mesh_.n_elx += 1

            if event.key in ['D', 'd']:
                self._mesh_.n_elx -= 1

            if event.key in ['up']:
                self._mesh_.Yc1 -= 1
                self._mesh_.Yc2 -= 1

            if event.key in ['down']:
                self._mesh_.Yc1 += 1
                self._mesh_.Yc2 += 1

            if event.key in ['left']:
                self._mesh_.Xc1 -= 1
                self._mesh_.Xc2 -= 1

            if event.key in ['right']:
                self._mesh_.Xc1 += 1
                self._mesh_.Xc2 += 1

            try:
                self._mesh_.gen_node_positions()
                render_mesh()
                pass

            except:
                pass

        def print_instructions():
            print(
                'Use arraow keys to move mesh, W and X to change refinement in  Y-directions, A and D to change refinement in X-direction')

        def ok(event):
            plt.close()

        plt.ioff()
        fig = plt.figure()

        # Doing some layout with subplots:
        fig.subplots_adjust(0.05, 0.05, 0.98, 0.98, 0.1)
        overview = plt.subplot2grid((12, 4), (0, 0), rowspan=11, colspan=4)

        overview.imshow(self.image, cmap=plt.cm.gray)

        data, = overview.plot([], [], 'ro')
        overview.autoscale(1, 'both', 1)

        but_ax1 = plt.subplot2grid((12, 4), (11, 2), colspan=1)
        ok_button = Button(but_ax1, 'OK')
        ok_button.on_clicked(ok)

        but_ax2 = plt.subplot2grid((12, 4), (11, 3), colspan=1)
        reset_button = Button(but_ax2, 'Reset')

        rectangle = RectangleSelector(overview, line_select_callback,
                                      drawtype='box', useblit=True,
                                      button=[1, 3],  # don't use middle button
                                      minspanx=5, minspany=5,
                                      spancoords='pixels')

        fig.canvas.mpl_connect('key_press_event', toggle_selector)

        _widgets = [rectangle, reset_button, ok_button]
        print_instructions()

        plt.show(block=True)

    def mesh(self, images, Xc1=0.0, Xc2=100.0, Yc1=0.0, Yc2=100., n_elx=4, n_ely=4, GUI=True, **kwargs):
        if isinstance(images, (ImageStack)):
            self.image = images[0]
        else:
            raise TypeError("Images should be in an ImageReader instance")

        if not type(Xc1) == float and type(Xc2) == float and type(Yc1) == float and type(Yc2) == float:
            raise TypeError("Coordinates should be given as floats")

        if not type(n_elx) == int and type(n_ely) == int:
            raise TypeError("Coordinates should be given as floats")

        element = BSplineSurface(self.deg_e, self.deg_n, **kwargs)

        self._mesh_ = Mesh(element, Xc1, Xc2, Yc1, Yc2, n_elx, n_ely)

        if GUI:
            self.__gui__()

        return copy(self._mesh_)


class Mesh(object):
    def __init__(self, element, corner1_x, corner2_x, corner1_y, corner2_y, n_elx, n_ely):
        """
        Mesh class

        Generates a grid based on the provided Finite Element definitions and geometrical measures.
        The class contains methods for generating the grid and for moving and resizing the grid.

         Parameters
         ----------
        element : object
            Instance of FiniteElement containing element definitions
        Xc1 : float
            X-Coordinate of upper left corner
        Yc1 : float
            Y-Coordinate of upper left corner
        Xc2 : float
            X-Coordinate of lower right corner
        Yc2 : float
            Y-Coordinate of lower right corner
        n_elx : int
            Number of elements in the x-direction
        n_ely : int
            Number of elements in the y-direction
         Returns
         -------
        Mesh :  Mesh object
         """
        self.element_def = element

        self.Xc1 = corner1_x
        self.Xc2 = corner2_x
        self.Yc1 = corner1_y
        self.Yc2 = corner2_y

        self.n_elx = n_elx
        self.n_ely = n_ely

        # Fields that are set after gen_mesh is called
        self.xnodes = None
        self.ynodes = None
        self.n_nodes = None
        self.n_elms = None

        self.gen_node_positions()

    def gen_node_positions(self):
        logger = logging.getLogger(__name__)
        try:
            self.xnodes, self.ynodes = make_grid(self.Xc1, self.Yc1, self.Xc2, self.Yc2,
                                                 self.n_elx,
                                                 self.n_ely, self.element_def)

            logger.info('Element contains %.1f X %.1f pixels and is divided in %i X %i ' % (
                (self.Xc2 - self.Xc1) / self.n_elx, (self.Yc2 - self.Yc1) / self.n_ely, self.n_elx, self.n_ely))

            self.n_nodes = len(self.xnodes)
            self.n_elms = 1

        except Exception as e:
            logger.exception("Mesh generation failed")

    def scale_mesh_y(self, factor):
        """
        Scale mesh in the y direction by a factor


         Parameters
         ----------
        factor : float
            The factor which the mesh is scaled by in the y direction

         """
        center = (self.Yc2 + self.Yc1) / 2.
        height = self.Yc2 - self.Yc1
        self.Yc1 = center + (height / 2.) * factor
        self.Yc2 = center - (height / 2.) * factor

    def scale_mesh_x(self, factor):
        """
        Scale mesh in the x direction by a factor


         Parameters
         ----------
        factor : float
            The factor which the mesh is scaled by in the x direction

         """
        center = (self.Xc2 + self.Xc1) / 2.
        height = self.Xc2 - self.Xc1
        self.Xc1 = center + (height / 2.) * factor
        self.Xc2 = center - (height / 2.) * factor

    def center_mesh_at(self, center_point_x, center_point_y):
        """
        Center the mesh at coordinates


         Parameters
         ----------
        center_pointx : float
            The center point of the mesh in the x-direction
        center_pointy : float
            The center point of the mesh in the y-direction
         """
        width = self.Xc2 - self.Xc1
        height = self.Yc2 - self.Yc1
        self.Xc1 = center_point_x - (width / 2.)
        self.Xc2 = center_point_x + (width / 2.)
        self.Yc1 = center_point_y - (height / 2.)
        self.Yc2 = center_point_y + (height / 2.)

    def single_element_mesh(self):
        """
        Convert mesh to a single element mesh
        """
        self.n_elx = self.element_def.degree_e + 1
        self.n_ely = self.element_def.degree_n + 1
        self.n_elms = 1
        self.gen_node_positions()





def mesh_translator(org_mesh, target_mesh, dic_results):
   """
   Mesh translator maps the nodal position history obtained with org_mesh to the corresponding
   nodal position history for target_mesh.
   The intended use of this function is to allow for a result set to be used as initial conditions for another analysis.
   TODO: This implementation currently only supports org_mesh being a single spline element.
   TODO: Write tests
   :param org_mesh: mesh instance
   :param target_mesh: mesh instance
   :param dic_results: dic_results instance
   :return: xnodesT,ynodeT corresponding to target_mesh
   """
   if org_mesh.ele.shape[1] is not 1:
       raise IOError("The origin mesh has to be a single element mesh!")
   if org_mesh.ynodes.shape[0] is not dic_results.xnodesT.shape[0]:
       raise IOError("The node-positions does not correspond to the original mesh")
   # We need the element coordinates to the nodes
   # We do this by using the original and target mesh in its un-deformed state
   es, ns, _, _ = find_element_coordinates(org_mesh.xnodes[org_mesh.ele[:, 0]],
                                           org_mesh.ynodes[org_mesh.ele[:, 0]], org_mesh.element_def,
                                           Xx=target_mesh.xnodes.flatten(),
                                           Yy=target_mesh.ynodes.flatten())
   node_x = np.dot(target_mesh.element_def.Nn(es, ns), dic_results.xnodesT[org_mesh.ele[:, 0], :])
   node_y = np.dot(target_mesh.element_def.Nn(es, ns), dic_results.ynodesT[org_mesh.ele[:, 0], :])
   return node_x, node_y


name = "muDIC"
from .solver import DICInput
from muDIC.post.viz import Fields, Visualizer
from muDIC.solver.correlate import DICInput, DICOutput
from . import IO
from . import elements
from . import filtering
from . import mesh
from . import mesh
from . import post
from . import vlab
from . import utils
from .IO import image_stack_from_list, image_stack_from_folder, ImageStack
from .mesh import Mesher
from .solver import *



import logging
from functools import partial

import numpy as np
import scipy.ndimage as nd
from scipy import optimize


def num_diff(xs, ys, func, component=(1, 1)):
    xs_mapped, ys_mapped = func(xs, ys)
    func_vals_x_grad_x = np.gradient(xs_mapped, axis=1)
    func_vals_x_grad_y = np.gradient(xs_mapped, axis=0)

    func_vals_y_grad_x = np.gradient(ys_mapped, axis=1)
    func_vals_y_grad_y = np.gradient(ys_mapped, axis=0)

    if component == (1, 1):
        return func_vals_x_grad_x  # / xs_spacing_x
    if component == (2, 1):
        return func_vals_x_grad_y
    if component == (1, 2):
        return func_vals_y_grad_x
    if component == (2, 2):
        return func_vals_y_grad_y
    else:
        raise ValueError("Incompatible component request")


def inverse(xs, ys, function, tol=1e-6):
    """Calculates the inverse of a displacement function

    Returns the arguments corresponding to the function values given.
    This is done by a Newton solver which minimizes:

    g(X) = X + f(X,Y) - x = 0
    where x is the coordinate of the displacement material point, having the coordinate X in its undeformed state.
    The coordinate Y is kept constant thoughout the optimization, and the routine is therefore not guarantied to
    converge.

    Parameters
    ----------
    xs : 2d-Numpy array
        x-coordinates
    ys : 2d-Numpy array
        x-coordinates
    function : func
        The displacement function
    tol : float
        The tolerance which the results are checked against

    Returns
    -------
    X,Y
        The coordinates in the undeformed state
    """
    logger = logging.getLogger()

    n_coords = xs.shape

    def f_x(x, y, a_x):
        u_x, _ = function(x, y, img_shape=n_coords)
        return x + u_x - a_x

    def f_y(y, x, a_y):
        _, u_y = function(x, y, img_shape=n_coords)
        return y + u_y - a_y

    x_deformed = xs.flatten()
    x_undeformed_guess = x_deformed.copy()

    y_deformed = ys.flatten()
    y_undeformed_guess = y_deformed.copy()

    x = optimize.newton(f_x, x_undeformed_guess, args=(y_undeformed_guess, x_deformed,))
    y = optimize.newton(f_y, y_undeformed_guess, args=(x_undeformed_guess, y_deformed,))

    error_x = np.max(np.abs(f_x(np.array(x), y_undeformed_guess, x_deformed)))
    error_y = np.max(np.abs(f_y(np.array(y), x_undeformed_guess, y_deformed)))

    if max(error_x, error_y) > tol:
        raise ValueError("The displacement function was not inverted successfully")

    logger.info("Larges error in y is: %f" % error_y)
    logger.info("Larges error in x is: %f" % error_x)

    return np.reshape(x, xs.shape), np.reshape(y, ys.shape)


class ImageDeformer(object):
    def __init__(self, coordinate_mapper, multiplicative,order=4):
        """Image deformer

        Deforms an image according to the settings given at instantiation.


        Parameters
        ----------
        coordinate_mapper : func
            A function taking two coordinates and returns two mapped coordinates
        multiplicative : bool
            Whether the coordinate mapping is multiplicative or not. Use True for deformation gradients
            and use False for displacement fields.
        order : int
            The interpolation order used

        Example
        -------
        Make a ImageDeformed which deforms the image according to a deformation gradient.
        NOTE: You should rather use the factory imageDeformer_from_defGrad.

        >>> import numpy as np
        >>> from muDIC import vlab
        >>> F = np.array([[1.1,.0], [0., 1.0]], dtype=np.float64)
        >>> coordinate_mapper = vlab.image_deformer.CoordinateMapper(partial(map_coords_by_defgrad, F=F))
        >>> vlab.image_deformer.ImageDeformer(coordinate_mapper,multiplicative=True)

        NOTE
        ----
        The image deformed returns a list of deformed images when called, the first being undeformed

        """
        self.multiplicative = multiplicative
        self.coodinate_mapper = coordinate_mapper
        self.order = order

    def __call__(self, img, steps=2):
        n, m = np.shape(img)
        def_imgs = []

        xn, yn = np.meshgrid(np.arange(n), np.arange(m))
        xn_mapped, yn_mapped = xn, yn

        for i in range(steps):
            if i == 0:
                Ic = img
            else:
                if self.multiplicative:
                    xn_mapped, yn_mapped = self.coodinate_mapper(xn_mapped, yn_mapped)
                else:
                    xn_mapped, yn_mapped = self.coodinate_mapper(xn, yn)
                    xn_mapped, yn_mapped = float(i) * (xn_mapped - xn) + xn, float(i) * (
                                yn_mapped - yn) + yn

                Ic = nd.map_coordinates(img, np.array([yn_mapped, xn_mapped]), order=self.order, prefilter=True, cval=0)

            def_imgs.append(Ic)

        return def_imgs


def map_coords_by_defgrad(xs, ys, F):
    """Maps coordinates based on a deformation gradient
    NOTE. The coordinates are first centered, implying that a rotation is about the middle of the image.

    Parameters
    ----------
    xs : 2d-Numpy array
        X-coordinates
    ys : 2d-Numpy array
        Y-coordinates
    F : 2x2-Numpy array
        The deformation gradient

    Returns
    -------
    xs_new,ys_new
        The mapped coordinates in the same ordering as input
    """
    # Center around x and y
    x_mid = np.float(xs.max()) / 2.
    y_mid = np.float(ys.max()) / 2.

    xs_centered = xs - x_mid
    ys_centered = ys - y_mid
    coords = np.array([xs_centered, ys_centered])
    F_inv = np.linalg.inv(F)

    coords_new = np.einsum('ij,jnm->inm', F_inv, coords)
    return coords_new[0, :, :] + x_mid, coords_new[1, :, :] + y_mid


def imageDeformer_from_defGrad(F):
    """Makes a ImageDeformer object from a deformation gradient

    Parameters
    ----------
    F : 2x2-ndarray
        The components of the deformation gradient


    Returns
    -------
    imageDeformer
        An ImageDeformer object which deformes images according to F

    Examples
    --------
    First, we define the deformation gradient and then make an ImageDeformer object:
        >>> import numpy as np
        >>> from muDIC import vlab
        >>> F = np.array([[1.1,.0], [0., 1.0]], dtype=np.float64)
        >>> image_deformer = vlab.imageDeformer_from_defGrad(F)

    """
    return ImageDeformer(partial(map_coords_by_defgrad, F=F), multiplicative=True)


def imageDeformer_from_uFunc(uFunc, **kwargs):
    """Makes a ImageDeformer object from a displacement function

    Parameters
    ----------
    uFunc : func
        The displacement function

    Returns
    -------
    imageDeformer
        An ImageDeformer object which deformes images according to the displacement function

    Examples
    --------
    First, we define the deformation gradient and then make an ImageDeformer object:
        >>> from muDIC import vlab
        >>> image_deformer = vlab.imageDeformer_from_uFunc(vlab.deformation_fields.bilat_harmonic,omega=4*np.pi)

    """
    coordinate_generator = partial(uFunc, **kwargs)
    coordinate_mapper_inv = partial(inverse, function=coordinate_generator)
    return ImageDeformer(coordinate_mapper_inv, multiplicative=False)



from __future__ import absolute_import

import logging
from copy import copy

import numpy as np
import scipy.ndimage as nd

from .reference import generate_reference as generate_reference
from ..IO.image_stack import ImageStack
from ..elements.fieldInterpolator import FieldInterpolator
from ..mesh.meshUtilities import Mesh
from ..solver.custom_exceptions import DidNotConverge
from ..utils import convert_to_img_frame, find_element_borders


def correlate_img_to_ref(node_pos, mesh, img, ref, settings):
    """
   Correlate an image to a reference
    The routine identifies the part of the image covered by the mesh and
    tries to perform image correlation on this part of the image.

    Parameters
    ----------
    node_pos : ndarray
       The position of the nodes
    mesh : Mesh
       The mesh object
    img : ndarray
       2d array containing the image
    ref : Reference
       The reference object
    settings : DICInput
       The settings which will be used during the analysis
   Returns
   -------
   updated node positions, current pixel values

   NOTES
   -------
   The function extracts a rectangular region of the image covered by the element, which may be very large
   if the mesh is tilted. This would reduce the performance of the routine
    """
    element_borders = find_element_borders(node_pos, mesh)

    image_frame, node_pos_img_coords = convert_to_img_frame(img, node_pos, mesh, element_borders, settings)

    node_position_increment, Ic = correlate_frames(node_pos_img_coords, mesh, image_frame, ref, settings)

    node_position_new = node_pos + node_position_increment
    return node_position_new, Ic


def correlate_frames(node_pos, mesh, img, ref, settings):
    """
    Parameters
    ----------
    node_pos : ndarray
       The position of the nodes
    mesh : Mesh
       The mesh object
    img : ndarray
       2d array containing the image frame
    ref : Reference
       The reference object
    settings : DICInput
       The settings which will be used during the analysis
   Returns
   -------
   updated node positions, current pixel values
    """

    logger = logging.getLogger(__name__)

    node_pos = np.copy(node_pos).astype(settings.precision)

    # Declare empty arrays
    pixel_pos = np.zeros((2, ref.n_pixels), dtype=settings.precision)
    dnod_x = np.zeros(mesh.n_nodes * 2)

    image_filtered = nd.spline_filter(img, order=3).transpose()

    for it in range(settings.maxit):

        # Find nodal positions within ROI
        np.dot(node_pos, ref.Nref_stack, out=pixel_pos)

        # Find pixel values for current coordinates
        Ic = nd.map_coordinates(image_filtered, pixel_pos, order=3, prefilter=False)

        # Calculate position increment as (B^T B)^-1 * (B^T*dIk) "Least squares solution"
        dnod = np.dot(ref.K, ref.I0_stack - Ic)

        # Add increment to nodal positions
        node_pos[0, :] += dnod[:mesh.n_nodes]
        node_pos[1, :] += dnod[mesh.n_nodes:]

        dnod_x += dnod

        # Check for convergence
        if np.max(np.abs(dnod)) < settings.tol:
            logger.info('Frame converged in %s iterations', it)
            return np.array((dnod_x[:mesh.n_nodes], dnod_x[mesh.n_nodes:])), Ic

        # Reset array values
    raise DidNotConverge('No convergence')


def store_stripped_copy(reference_generator, storage):
    """
   Store a stripped copy of the reference object
    The stored reference object is stripped from all resource intensive variables

   Parameters
   ----------
   reference_generator : func
       The reference generator function
   storage : list
       The list where the stripped references are appended
   Returns
   -------
   reference_generator
    """

    def wrapper(*args, **kwargs):
        ref = reference_generator(*args, **kwargs)
        ref_light = copy(ref)

        # Remove the heavy fields
        ref_light.Nref_stack = None
        ref_light.B_stack = None
        ref_light.K = None
        storage.append(ref_light)
        return ref

    return wrapper


def correlate(inputs):
    """
   Main correlation routine
    This routine manages result storage, reference generation and
     the necessary logic for handling convergence issues.

   Parameters
   ----------
   inputs : DIC_input object
       The input object containing all necessary data for performing a DIC analysis.

   Returns
   -------
   node_coords, reference_stack, Ic_stacks
    """

    logger = logging.getLogger(__name__)

    mesh = inputs.mesh
    images = inputs.images
    settings = inputs

    # Do the initial setup

    images.image_reader.precision = settings.precision

    Ic_stacks = list()
    reference_stack = list()
    node_position_t = list()

    if settings.store_internals:
        gen_ref = store_stripped_copy(generate_reference, storage=reference_stack)
    else:
        gen_ref = generate_reference

    if settings.node_hist:
        node_coords = np.array(settings.node_hist, dtype=settings.precision)[:, :, 0]
    else:
        node_coords = np.array((mesh.xnodes, mesh.ynodes), dtype=settings.precision)

    reference = gen_ref(node_coords, mesh, images[0], settings, image_id=0)

    # Correlate the image frames

    try:
        for image_id in range(settings.max_nr_im):
            logger.info('Processing frame nr: %i', image_id)

            if settings.node_hist:
                node_coords = np.array(settings.node_hist, dtype=settings.precision)[:, :, image_id]

            if image_id in settings.ref_update:
                logger.info('Updating reference at %i', image_id)
                reference = gen_ref(node_coords, mesh, images[image_id - 1], settings, image_id=(image_id - 1))

            img = images[image_id]

            try:
                node_coords, Ic = correlate_img_to_ref(node_coords, mesh, img, reference, settings)

            except DidNotConverge:
                # If the reference is new, there is nothing more to do. Otherwise, update the reference and try again.
                if reference.image_id == image_id - 1:
                    logger.info('Correlation failed at frame %i', image_id)
                    break
                else:
                    logger.info('Updating reference at %i', image_id)
                    reference = gen_ref(node_coords, mesh, images[image_id - 1], settings, image_id - 1)
                    node_coords, Ic = correlate_img_to_ref(node_coords, mesh, img, reference, settings)

            except Exception as e:
                logger.info('Correlation failed with error at frame %i', image_id)
                logger.exception(e)
                break

            if settings.store_internals:
                Ic_stacks.append(Ic)

            node_position_t.append(node_coords)

    finally:
        return np.array(node_position_t), reference_stack, Ic_stacks


class DICAnalysis(object):

    def __init__(self, inputs):

        """
         DIC analysis

        The analysis object verifies and stores the DIC_input object.
        When instantiated, the .run() method can be called, initiating the DIC analysis.

         Parameters
         ----------
         inputs : DIC_input object
             The input object containing all neccessary data for performing a DIC analysis.

         Returns
         -------
         DIC_analysis object


         Examples
         --------
        The following example runs a virtual experiment

            >>> import muDIC as dic
            >>> import numpy as np
            >>> import muDIC.vlab as vlab

            >>> image_shape = (2000, 2000)
            >>> speckle_image = vlab.rosta_speckle(image_shape, dot_size=4, density=0.32, smoothness=2.0, layers=4)

            >>> F = np.array([[1.1, .0], [0., 1.0]], dtype=np.float64)
            >>> image_deformer = vlab.imageDeformer_from_defGrad(F)
            >>> downsampler = vlab.Downsampler(image_shape=image_shape, factor=4, fill=0.8, pixel_offset_stddev=0.1)
            >>> noise_injector = vlab.noise_injector("gaussian", sigma=.1)
            >>> image_generator = vlab.VirtualExperiment(speckle_image=speckle_image, image_deformer=image_deformer,
            >>>                              downsampler=downsampler, noise_injector=noise_injector, n=n)
            >>> image_stack = dic.ImageStack(image_generator)

            >>> mesher = dic.Mesher(deg_n=1,deg_e=1)
            >>> mesh = mesher.mesh(image_stack)

            >>> input = muDIC.solver.correlate.DIC_input(mesh, image_stack)
            >>> dic_job = dic.DIC_analysis(input)
            >>> results = dic_job.run()

         """
        self.logger = logging.getLogger()

        self.__input__ = self.__verify_dic_input__(inputs)

    def run(self):
        """
         Run analysis

         Parameters
         ----------

         Returns
         -------
         DIC_output object



         """
        node_x, node_y, reference_stack, Ic_stack = self.__solve__()

        return DICOutput(node_x, node_y, self.__input__, ref_stack=reference_stack, Ic_stack=Ic_stack)

    def get_input(self):
        return self.__input__

    def __solve__(self):
        node_position, reference_stack, Ic_Stack = correlate(self.__input__)
        # TODO: Remove the need of transposing the matrices
        return node_position[:, 0, :].transpose(), node_position[:, 1,
                                                   :].transpose(), reference_stack, Ic_Stack

    @staticmethod
    def __verify_dic_input__(inputs):
        """
        Input type checker. Verifies all input and calculates missing values if they can be deduced from the others.

        """
        inputs_checked = inputs

        if not isinstance(inputs_checked, DICInput):
            raise TypeError('Inputs has to be an instance of the DICInput class')

        if not isinstance(inputs_checked.images, (ImageStack)):
            raise TypeError('Image stack is not an instance of Image_reader')

        if not isinstance(inputs_checked.mesh, Mesh):
            raise TypeError('Mesh should be an instance of Mesh')

        if isinstance(inputs_checked.max_nr_im, int) and inputs_checked.max_nr_im <= len(inputs_checked.images):
            pass
        else:
            inputs_checked.max_nr_im = len(inputs_checked.images)

        if not isinstance(inputs_checked.mesh.element_def, FieldInterpolator):
            raise TypeError('Finite element should be an instance of Finite_Element')
        inputs_checked.elm = inputs_checked.mesh.element_def

        if not isinstance(inputs_checked.ref_update, (list, tuple)):
            raise TypeError('Reference update frames should be specified by a list or tuple')

        if type(inputs_checked.maxit) is not int:
            raise TypeError('Maximum number of iterations should be specified by an integer')

        if type(inputs_checked.pad) is not int:
            raise TypeError('Padding width should be specified by an integer')

        return inputs_checked


class DICInput(object):

    def __init__(self, mesh, image_stack, ref_update_frames=[50, 150], maxit=40, max_nr_im=None, pad=10,
                 store_internals=False, node_hist=None, precision="double", interpolation_order=3, block_size=1e7):

        """
         DIC output container

        This class contains all the necessary inputs for a DIC analysis.

         Parameters
         ----------
         mesh : Mesh object
             The mesh corresponding to the image stack
         image_stack : ImageStack
             The image_stack containing all images for the DIC analysis
         ref_update_frames : list, optional
             A list of indices for the images where a reference update should be performed
         maxit : int, optional
             The maximum allowed number of iterations for the DIC solver before a reference update is performed.
         max_nr_im : int, optional
             The maximum number of images to be analyses
         pad : int, optional
             The amount of padding used around the mesh when the image sub-frame is extracted
         store_internals : bool, optional
             If True, all references are stored and available in the DIC_output
         node_hist : ndarray, optional
             An array containing the nodal positions for each image frame. This is used as initial conditions for the solver
         precision : strings, optional
             The number precision to be used, either "single" or "double".
         interpolation_order : int, optional
             The polynomial order used for image interpolation during the DIC analysis.
         Returns
         -------
         DIC_input object


         Examples
         --------



         """

        self.mesh = mesh
        self.images = image_stack
        self.max_nr_im = max_nr_im
        self.ref_update = ref_update_frames
        self.maxit = maxit
        self.pad = pad
        self.elm = None
        self.tol = 1e-6
        self.store_internals = store_internals
        self.node_hist = node_hist
        self.interpolation_order = interpolation_order
        self.block_size = block_size

        if precision == "single":
            self.precision = np.float32
        else:
            self.precision = np.float64


class DICOutput(object):
    def __init__(self, node_x, node_y, settings, ref_stack=None, Ic_stack=None):
        """
        DIC output container

        This object contains the DIC results

         Parameters
         ----------
         node_x : ndarray
             The nodal x positions for each frame
         node_y : ImageStack
             The nodal y positions for each frame
         settings : list, optional
             The DIC_input object used for the analysis
         ref_stack : int, optional
             A list containing all the references used during the analysis
         Ic_stack : int, optional
             A list containing all the pixel intensities used during the analysis
         Returns
         -------
         DIC_output object
         """

        self.xnodesT = node_x
        self.ynodesT = node_y
        self.crosscorrelation = None
        self.reference = ref_stack
        self.Ic_stack = Ic_stack
        self.settings = settings
        self.settings.images = None



# -*- coding: utf-8 -*-
""" Down sampling module which can be used to downsample images by interpolation,
include sensor artifacts such as fillfactor and pixel location inaccuracies.

This module is directly based on the conference paper:

>>> Jean-José Orteu, Dorian Garcia, Laurent Robert, Florian Bugarin. A speckle-texture image generator.
    Slangen, P. Speckle06 - Speckles from grains to flowers, Sep 2006, Nimes, France. Extrait de: Speckle06
    - Speckles from grains to flowers - Proceedings of Spie / sous la dir. de P. SLANGEN et C. CERRUTTI.
    - ISBN 0-8194-6426-0, 6341, 2006, <10.1117/12.695280>. <hal-01644899>


"""

import numpy as np
from scipy import ndimage


def coord_subpos(coord, fillfactor, n, i, sigma):
    """
    Super sampled coordinates

    This function calculates the N sample coordinates with a given fillcator and in addition:
    - Inaccuracies in pixel position can be introduced by defining the standard deviation of a normal distribution
      which describes the pixel offset distribution.

    Parameters
    ----------
    coord : NxN array
        A coordinate array with the center coordinates about which the spread is calculated
    fill_factor : Float
        The area fraction covered by a pixel
    n : Int
        The number of super sampling points
    i : NxN array
        Index array for the super sampling points. Has the shape (n x n)
    permutation_sigma : Float
        The standard deviation of the normal distribution describing the pixel offset

    Returns
    -------
    Super sampled coordinates: N x N x n x n array


    Examples
    --------
    The following example bins the image in 4 x 4 bins

        >>> from muDIC import vlab
        >>> import numpy as np
        >>> pixel_centers = np.ones((1,1))
        >>> n=5
        >>> indices = np.arange(n)[:,np.newaxis] * np.ones((n,n))
        >>> vlab.downsampler.coord_subpos(pixel_centers, fillfactor=1.0, n, indices, sigma=0.):
    array([[[[0.6, 0.6, 0.6, 0.6, 0.6],
         [0.8, 0.8, 0.8, 0.8, 0.8],
         [1. , 1. , 1. , 1. , 1. ],
         [1.2, 1.2, 1.2, 1.2, 1.2],
         [1.4, 1.4, 1.4, 1.4, 1.4]]]])


    """
    distribution_mean = 0.
    perturbation = (np.sqrt(fillfactor) / n) * np.random.normal(distribution_mean, sigma)
    return coord[:, :, np.newaxis, np.newaxis] + np.sqrt(fillfactor) * (
            2. * i[np.newaxis, np.newaxis, :, :] + 1. - n) / (
                   2. * n) + perturbation


def chess_board(n):
    image_width = 2
    image_value = 2.0

    frame_with = 1
    frame_value = 1.0

    image = np.ones((image_width, image_width)) * image_value
    framed_image = np.pad(image, frame_with, mode="constant", constant_values=frame_value)

    tile = np.zeros((8, 8))
    tile[:4, :4] = framed_image
    tile[4:, :4] = framed_image * -1
    tile[:4, 4:] = framed_image * -1
    tile[4:, 4:] = framed_image

    tile_assembly = np.tile(tile, (n, n)).astype(np.float)

    return tile_assembly


class Downsampler(object):
    def __init__(self, image_shape, factor=4, fill=1.0, pixel_offset_stddev=0.0):
        """
        Downsampler for gray-scale images.

        This function downsamples a NxN image by an integer value with the following addition:
        - A fillfactor between 0 and 1 can be given in order to mimic a image sensors fillfactor
        - Inaccuracies in pixel position can be introduced by defining the standard deviation of a normal distribution
          which describes the pixel offset distribution.

        Parameters
        ----------
        image : NxN array
            The image which will be downsampled
        down_sampling_factor : Integer
            The downsampling factor
        fill_factor : Float
            The area fraction covered by a pixel
        permutation_sigma : Float
            The standard deviation of the normal distribution describing the pixel offset

        Returns
        -------
        downsampled_image: NxN array


        Examples
        --------
        The following example bins the image in 4 x 4 bins

        >>> image = np.ones((20,20))
            downsampled_image = downsample_image(image,4,fill_factor=1.0,permutation_sigma=0.0)
            downsampled_image.shape
        (5,5)


        """

        self.image_shape = image_shape
        self.factor = factor
        self.fill = fill
        self.pixel_offset_stddev = pixel_offset_stddev

        self.dim_n, self.dim_m = self.image_shape
        if self.dim_m != self.dim_n:
            raise ValueError("Only N x N images are supported, %i x %i was given" % (self.dim_m, self.dim_n))

        if divmod(self.dim_m, self.factor)[1] != 0:
            raise ValueError("Shape of image and downsampling factor does not match")

        if type(self.factor) is not int:
            raise TypeError("Down sampling factor has to be an integer")

        if self.fill > 1. or self.fill < 0.:
            raise ValueError("Fillfactor has to be between 1 and 0")

        self.coordinates = self.__generate_coordinates__()

    def __generate_coordinates__(self):

        n_subpoints = np.floor(self.factor * np.sqrt(self.fill)).astype(np.int)

        num_segments_n = np.float(self.dim_n) / self.factor
        num_segments_m = np.float(self.dim_m) / self.factor

        tile_i = np.arange(num_segments_n, dtype=np.float)
        tile_j = np.arange(num_segments_m, dtype=np.float)

        i, j = np.meshgrid(tile_i, tile_j)

        # Pertubate the pixels
        i_shift = np.random.normal(loc=0.0, scale=self.pixel_offset_stddev, size=i.shape)
        j_shift = np.random.normal(loc=0.0, scale=self.pixel_offset_stddev, size=j.shape)

        i += i_shift
        j += j_shift

        ind_i, ind_j = np.meshgrid(np.arange(n_subpoints, dtype=np.float), np.arange(n_subpoints, dtype=np.float))

        subframe_i = (coord_subpos(i, self.fill, n_subpoints, ind_i,
                                   self.pixel_offset_stddev) + .5) * np.float(
            self.factor) - .5

        subframe_j = (coord_subpos(j, self.fill, n_subpoints, ind_j,
                                   self.pixel_offset_stddev) + .5) * np.float(
            self.factor) - 0.5

        coordinates = np.zeros(
            (2, np.int(num_segments_n), np.int(num_segments_m), n_subpoints, n_subpoints))

        coordinates[1, :, :, :, :] = subframe_i
        coordinates[0, :, :, :, :] = subframe_j
        return coordinates

    def __call__(self, image):
        if image.shape[0] != self.dim_m or image.shape[1] != self.dim_n:
            raise ValueError("Invalid shape")

        downsampled_image_bins = ndimage.map_coordinates(image, self.coordinates, order=3, prefilter=True, cval=np.nan)

        return np.mean(downsampled_image_bins, axis=(2, 3))



from functools import reduce
import scipy.ndimage as nd
import numpy as np


def convert_to_img_frame(img, node_position, mesh, borders, settings):
    local_node_pos = np.zeros((2, mesh.element_def.n_nodes), dtype=settings.precision)

    # Partition image
    image_frame = extract_subframe(img, borders, settings.pad)

    # Determine nodal positions in image frame coordinates
    local_node_pos[0, :] = node_position[0] + settings.pad - borders[0, :]
    local_node_pos[1, :] = node_position[1] + settings.pad - borders[2, :]

    return image_frame, local_node_pos


def generate_edge_coordinates(seed):
    seeding = np.linspace(0., 1., seed)
    es, ns = np.meshgrid(seeding, seeding)
    mask = np.ones_like(es, dtype=np.bool)
    mask[1:-1, 1:-1] = 0
    return es[mask], ns[mask]


def find_element_borders(node_position, mesh, seed=20):
    e, n = generate_edge_coordinates(seed)
    N_at_borders = mesh.element_def.Nn(e.flatten(), n.flatten())

    # Find global coordinates of elements
    pixel_x = np.einsum("jk,k->j", N_at_borders, node_position[0])
    pixel_y = np.einsum("jk,k->j", N_at_borders, node_position[1])

    axis = None
    # [Xmin_Xmax,Ymin,Ymax,elm_nr]
    borders = np.zeros((4, mesh.n_elms), dtype=np.int)
    borders[0, :] = np.min(pixel_x, axis=axis)
    borders[1, :] = np.max(pixel_x, axis=axis)
    borders[2, :] = np.min(pixel_y, axis=axis)
    borders[3, :] = np.max(pixel_y, axis=axis)

    return borders


def extract_subframe(img, borders, pad):
    return img[borders[2, 0] - pad:borders[3, 0] + pad, borders[0, 0] - pad:borders[1, 0] + pad]


def find_borders(coord):
    return int(np.min(np.floor(coord))), int(np.max(np.ceil(coord)))


def find_inconsistent(ep, ny):
    rem1 = np.where(ep > 1.)
    rem2 = np.where(ep < 0.)
    rem3 = np.where(ny > 1.)
    rem4 = np.where(ny < 0.)
    return reduce(np.union1d, [rem1[0], rem2[0], rem3[0], rem4[0]])

def extract_points_from_image(image, coordinates):
    return nd.map_coordinates(image, coordinates, order=3, prefilter=True)

def image_coordinates(image):
    xs, ys = np.meshgrid(np.arange(image.shape[0]), np.arange(image.shape[1]))
    return xs,ys


import numpy as np
import scipy.ndimage as nd
from scipy.ndimage import median_filter


def highpass_gaussian(image, sigma=2.0):
    return image - nd.gaussian_filter(image, sigma=sigma)


def lowpass_gaussian(image, sigma=2.0):
    return nd.gaussian_filter(image, sigma=sigma)


def homomorphic_median(image, sigma=10):
    log_img = np.log(image)
    return np.exp(log_img - median_filter(log_img, int(sigma)))



import dill


def save(dic, name):
    try:
        with open(name + '.p', 'wb') as myfile:
            dill.dump(dic, myfile)
    except:
        raise IOError('Could not save to file')


def load(name):
    try:
        with open(name + '.p', 'rb') as myfile:
            return dill.load(myfile)
    except:
        raise TypeError("Invalid inputs")



import numpy as np


def harmonic_x(xs, _, amp=1.1, omega=2.0 * np.pi, img_shape=(500, 500)):
    """
    Displacement field being sinusoidal along the x-axis


    Parameters
    ----------
    xs : array
        The x-coodinates
    ys : array
        The y-coordinates
    omega : float
        The angular frequency
    img_shape : tuple
        The shape of the field


    Returns
    -------
    u_x,u_y : array
        The displacement field values in each direction

    """

    # Center around x and y
    xs = xs.astype(np.float)

    xs_scaled = xs / np.float(img_shape[0])

    xs_mapped = amp * np.sin(omega * xs_scaled)
    return xs_mapped, np.zeros_like(xs_mapped)


def harmonic_bilat(xs, ys, amp=1.1, omega=np.pi * 2., img_shape=(200, 200)):
    """
    Displacement field being sinusoidal along the both axes


    Parameters
    ----------
    xs : array
        The x-coodinates
    ys : array
        The y-coordinates
    omega : float
        The angular frequency
    img_shape : tuple
        The shape of the field


    Returns
    -------
    u_x,u_y : array
        The displacement field values in each direction

    """

    xs_scaled = xs / np.float(img_shape[0])
    ys_scaled = ys / np.float(img_shape[1])

    xs_mapped = amp * np.sin(omega * xs_scaled) * np.sin(omega * ys_scaled)
    ys_mapped = amp * np.sin(omega * xs_scaled) * np.sin(omega * ys_scaled)
    return xs_mapped, ys_mapped



import logging
import os
from copy import copy
from functools import partial

import numpy as np
from scipy import ndimage


class ImageStack(object):
    def __init__(self, image_reader, filter=None):
        """
        Image stack

        ImageStacks are responsible for handling image imports, filtering and ordering.


        Various functionality is available, such as:

        * Adding a filter
        * Skipping images
        * Reversing the order of the images

         Parameters
         ----------
         image_reader : ImageReader or ImageGetter
             An object which handles the import of the images
         filter : Filter
             A function which filters the image

         Examples
         ----------
        >>> import muDIC as dic
        >>> import numpy as np
        >>> image_list = [np.ones((3,3))*i for i in range(10)]
        >>> image_stack = dic.image_stack_from_list(image_list)
        >>> image_stack.set_filter(dic.filtering.lowpass_gaussian,sigma=2)
        >>> image_stack.skip_images(range(9))
        >>> image_stack[0]
            array([[9., 9., 9.],
            [9., 9., 9.],
            [9., 9., 9.]])

         """

        self.logger = logging.getLogger()
        self.image_reader = image_reader
        self._all_img_ids_ = list(range(len(self.image_reader)))
        self._active_img_ids_ = copy(self._all_img_ids_)
        if filter is None:
            self._filter_ = lambda img: img
        else:
            self._filter_ = filter

    def set_filter(self, filter, **kwargs):
        """
        Set a filter

        The filter is applied to all frames.
        The filter has to be a function which takes an image as input and returns an image.

         Parameters
         ----------
         filter : func
             A function which takes an image as input and returns an image
         kwargs :
             The arguments that should be used for the filter.
         """
        self._filter_ = partial(filter, **kwargs)

    def __getitem__(self, index):
        return self._filter_(self.image_reader(self._active_img_ids_[index]))

    def revere_order(self):
        """
        Reverse the order of the images in the stack

        """
        self._active_img_ids_ = self._active_img_ids_[::-1]
        self.logger.info("Reversing the order of the image stack")

    def __len__(self):
        return len(self._active_img_ids_)

    def skip_images(self, skip_frames):
        """
        Skip the frames with the ids listed in skip_frames

         Parameters
         ----------
         skip_frames : list or tuple
             The list of frame ids to be skipped
         """

        if type(skip_frames) is not list and type(skip_frames) is not tuple:
            raise TypeError('Frames has to be given as a list or tuple with integers')

        elif not all((type(ind) == int) for ind in skip_frames):
            raise TypeError('Frame ids have to be integers')

        elif max(skip_frames) > len(self._all_img_ids_) or min(skip_frames) < 0:
            raise ValueError('Frame id outside bounds')

        self._active_img_ids_ = [frame for frame in self._active_img_ids_ if frame not in skip_frames]
        self.logger.info("Skipping frames. Length of image stack is now %i" % len(self._active_img_ids_))

    def use_every_n_image(self, n):

        self._active_img_ids_ = self._active_img_ids_[::n]
        self.logger.info("Using every %i frame. Length of image stack is now %i" % (n, len(self._active_img_ids_)))


class ImageReader(object):
    def __init__(self, image_paths):
        """
        Image reader which reads files from HD

        When the ImageReader is called, it returns an image corresponding to the index.

         Parameters
         ----------
         image_paths : list
             A list of paths to images.

         """
        self._image_paths_ = image_paths
        self.precision = np.float64

    def __len__(self):
        return len(self._image_paths_)

    def __call__(self, index, rotate=False):
        if not rotate:
            return ndimage.imread(self._image_paths_[index], flatten=True).astype(self.precision)
        else:
            return ndimage.rotate(ndimage.imread(self._image_paths_[index], flatten=True).astype(self.precision),
                                  rotate)


class ImageListWrapper(object):
    def __init__(self, images):
        """
        Image getter which wraps a list of images

        When the ImageGetter is called, it returns an image corresponding to the index.

         Parameters
         ----------
         images : list
             A list of images as numpy.ndarray

         """
        self._images_ = images
        self.precision = np.float64

    def __len__(self):
        return len(self._images_)

    def __call__(self, index, rotate_ang=False):
        if not rotate_ang:
            return self._images_[index].astype(self.precision)
        else:
            return ndimage.rotate(self._images_[index].astype(self.precision), angle=rotate_ang)


def find_file_names(path, type=".png"):
    """
    Finds all files with the given extension in the folder path.

     Parameters
     ----------
     path : str
         The path to the folder containing the files of interest
     type : str
         The file postfix such as ".png", ".bmp" etc.

     Returns
     -------
    List of filenames

     """
    return sorted([path + file for file in os.listdir(path) if file.endswith(type)])


def image_stack_from_folder(path_to_folder, file_type='.png'):
    """
    Make an ImageStack containing the images within a folder


     Parameters
     ----------
     path_to_folder : str
         The path to the folder containing the images of interest
     file_type : str
         The file postfix such as ".png", ".bmp" etc.

     Returns
     -------
     ImageStack object


     Examples
     --------
    >>> path_to_folder = r"/the/path/to/the/images/"
    >>> image_stack = image_stack_from_folder(path_to_folder,file_type=".bmp")
    ImageStack
     """
    logger = logging.getLogger()
    supported_filetypes = ['.png', '.bmp', '.tif']

    if type(path_to_folder) not in [str]:
        raise TypeError('Path has to be a string')

    if type(file_type) is not str or file_type not in supported_filetypes:
        # TODO: include more valid file types
        raise TypeError('Filetype has to be: %s' % str.join(*supported_filetypes))

    file_names_all = find_file_names(path_to_folder, file_type)
    logger.info("Found %i images in folder" % len(file_names_all))

    return ImageStack(ImageReader(file_names_all))


def image_stack_from_list(image_list):
    """
    Make an ImageStack containing the images in the list


     Parameters
     ----------
     image_list : list or tuple
         The list of images as numpy.ndarray

     Returns
     -------
     ImageStack object


     Examples
     --------
        >>> import muDIC as dic
        >>> import numpy as np
        >>> image_list = [np.random.rand(10,10) for i in range(10)]
        >>> image_stack = dic.image_stack_from_folder(image_list)
    ImageStack
     """
    logger = logging.getLogger()

    if type(image_list) not in [list, tuple]:
        raise TypeError('image_list has to be either a list or tuple')

    if type(image_list[0]) is not np.ndarray:
        raise TypeError("Images has to be numpy.ndarrays")

    logger.info("Found %i images in list" % len(image_list))

    return ImageStack(ImageListWrapper(image_list))



import logging

import numpy as np
from numba import jit, prange

from .fieldInterpolator import FieldInterpolator


# TODO: This thing should be refactored such that it runs i parallel

class BSplineSurface(FieldInterpolator):
    def __init__(self, deg_e=3, deg_n=3, n_ctrlpts_e=6, n_ctrlpts_n=6):
        """
        B-spline surface

        This B-spline surface is based on algorithms found in:

        >>> "The NURBS book" by Les Piegl and Wayne Tiller ISBN: 9783540615453

        The definition is made such that the Basis-functions are evaluated first allowing for evaluation of the
        surface by a simple dot product of the basis functions and the control point coordinates without recalculating
        the basis functions.


         Parameters
         ----------
         deg_e : int
             The order of the basis function in the e-direction
         deg_n : int
             The order of the basis function in the n-direction
         n_ctrlpts_e : int
             The number of control points in the e-direction
         n_ctrlpts_n : int
             The number of control points in the n-direction
         Returns
         -------
         B-spline surface object


         Examples
         --------
            >>> import muDIC as dic
            >>> import numpy as np
            >>> es,ns = np.meshgrid(np.linspace(0.,1.,10),np.linspace(0.,1.,10))
            >>> surface = dic.elements.B_spline_surface(deg_e=3, deg_n=3, n_ctrlpts_e=6, n_ctrlpts_n=6)
            >>> bases = surface.Nn(es.flatten(),ns.flatten())
         """

        self.logger = logging.getLogger()

        if isinstance(deg_e, int) and isinstance(deg_n, int):
            self.degree_e = deg_e
            self.degree_n = deg_n

        if isinstance(n_ctrlpts_e, int) and isinstance(n_ctrlpts_n, int):
            self.n_nodes_e = np.int64(n_ctrlpts_e)
            self.n_nodes_n = np.int64(n_ctrlpts_n)

        self.__gen_uniform_knotvectors__()

        self._indices_ = self._make_ctrlpt_indices_()
        self.ctrl_pos_e, self.ctrl_pos_n = self._make_grid_()
        self.n_nodes = self.n_nodes_e * self.n_nodes_n

    def _update_internals_(self):
        self.__gen_uniform_knotvectors__()
        self._indices_ = self._make_ctrlpt_indices_()
        self.ctrl_pos_e, self.ctrl_pos_n = self._make_grid_()
        self.n_nodes = self.n_nodes_e * self.n_nodes_n

    def _make_ctrlpt_indices_(self):
        return np.arange(self.n_nodes_e * self.n_nodes_n).reshape((self.n_nodes_e, self.n_nodes_n))

    def set_n_nodes(self, grid_shape):
        n_nodes_e, n_nodes_n = grid_shape

        if n_nodes_e < self.degree_e + 1 or n_nodes_e < self.degree_e + 1:
            raise ValueError("The number of nodes along an axis has to be larger than the degree of the polynomial +1")

        self.n_nodes_e = np.int64(n_nodes_e)
        self.n_nodes_n = np.int64(n_nodes_n)

        self._update_internals_()

    def set_degree(self, degs):
        dege, degn = degs
        self.degree_e = np.int64(dege)
        self.degree_n = np.int64(degn)

        self._update_internals_()

    def Nn(self, es, ns):
        """
         Evaluate the B-spline surface basis functions at the given coordinates in [0,1].
         This algorithm corresponds to A3.5 but without the multiplication with the nodal coordinates.

         >>> "The NURBS book" by Les Piegl and Wayne Tiller ISBN: 9783540615453

        The order to evaluate the the surface for a given set of control points, np.dot(Nn(e,n), ctrlpts)
        has to be evaluated.


          Parameters
          ----------
          es : ndarray
              1d array, with the coordinates to be evaluated the e-direction
          ns : ndarray
              1d array, with the coordinates to be evaluated the n-direction

          Returns
          -------
          basis_funcs : ndarray
                2D array, with the evaluated basis functions.
                The shape of the matrix is size(es) x num_crtl_pts

        """

        if es.ndim != 1 or ns.ndim != 1:
            raise ValueError("Only 1d arrays are accepted")

        if ns.max() > 1. or ns.min() < 0. or es.max() > 1. or es.min() < 0.:
            raise ValueError("B-spline coordinates has to be within [0,1]")

        spans_e = self._find_span_array_(self.degree_e, self._knotvector_e_, self.n_nodes_e, ns)

        bases_e = self.basis_functions_array(self.degree_e, self._knotvector_e_, spans_e, ns)

        spans_n = self._find_span_array_(self.degree_n, self._knotvector_n_, self.n_nodes_n, es)

        bases_n = self.basis_functions_array(self.degree_n, self._knotvector_n_, spans_n, es)

        results = self.__partial_outer_product__(self.n_nodes_e, self.n_nodes_n, len(es), spans_n, bases_n, spans_e,
                                                 bases_e,
                                                 self.degree_n,
                                                 self.degree_e, self._indices_)

        return results

    def dxNn(self, vs, us):
        """
         Evaluate the B-spline surface basis function derivatives in X at the given coordinates in [0,1].
         This algorithm corresponds to A3.5 but without the multiplication with the nodal coordinates.

         >>> "The NURBS book" by Les Piegl and Wayne Tiller ISBN: 9783540615453

        The order to evaluate the the surface for a given set of control points, np.dot(Nn(e,n), ctrlpts)
        has to be evaluated.


          Parameters
          ----------
          es : ndarray
              1d array, with the coordinates to be evaluated the e-direction
          ns : ndarray
              1d array, with the coordinates to be evaluated the n-direction

          Returns
          -------
          basis_funcs : ndarray
                2D array, with the evaluated basis function derivatives.
                The shape of the matrix is size(es) x num_crtl_pts

        """
        return self.__dKNn__(vs, us, 0, 1)

    def dyNn(self, vs, us):
        """
         Evaluate the B-spline surface basis function derivatives in Y at the given coordinates in [0,1].
         This algorithm corresponds to A3.5 but without the multiplication with the nodal coordinates.

         >>> "The NURBS book" by Les Piegl and Wayne Tiller ISBN: 9783540615453

        The order to evaluate the the surface for a given set of control points, np.dot(Nn(e,n), ctrlpts)
        has to be evaluated.


          Parameters
          ----------
          es : ndarray
              1d array, with the coordinates to be evaluated the e-direction
          ns : ndarray
              1d array, with the coordinates to be evaluated the n-direction

          Returns
          -------
          basis_funcs : ndarray
                2D array, with the evaluated basis function derivatives.
                The shape of the matrix is size(es) x num_crtl_pts

        """

        return self.__dKNn__(vs, us, 1, 0)

    def __dKNn__(self, vs, us, k, l):

        du = 1
        dv = 1

        n_pts = np.min([np.size(us), np.size(vs)]).astype(np.int64)

        spans_e = self._find_span_array_(self.degree_e, self._knotvector_e_, self.n_nodes_e, us)

        bases_ders_e = self.__basis_functions_ders_array__(self.degree_e, self._knotvector_e_, spans_e, us, du)

        span_n = self._find_span_array_(self.degree_n, self._knotvector_n_, self.n_nodes_n, vs)

        bases_ders_n = self.__basis_functions_ders_array__(self.degree_n, self._knotvector_n_, span_n, vs, dv)

        results_ordered = self.__partial_outer_product__(self.n_nodes_n, self.n_nodes_e, n_pts, span_n,
                                                         bases_ders_n[:, l, :],
                                                         spans_e,
                                                         bases_ders_e[:, k, :], self.degree_n, self.degree_e,
                                                         self._indices_)

        return results_ordered

    def _make_grid_(self):

        es = np.linspace(0., 1., self.n_nodes_e)
        ns = np.linspace(0., 1., self.n_nodes_n)

        ctrlpts_epos, ctrlpts_npos = np.meshgrid(ns, es)

        return ctrlpts_epos.flatten(), ctrlpts_npos.flatten()

    def __gen_uniform_knotvectors__(self):
        self._knotvector_e_ = np.array(self._gen_uniform_knotvector_(self.degree_e, self.n_nodes_e))
        self._knotvector_n_ = np.array(self._gen_uniform_knotvector_(self.degree_n, self.n_nodes_n))

    @staticmethod
    @jit(nopython=True)
    def _find_span_(degree, knotvector, num_ctrlpts, knot, tol=1e-5):

        """ Algorithm A2.1 of The NURBS Book by Piegl & Tiller."""
        # Number of knots; m + 1
        # Number of control points; n + 1
        # n = m - p - 1; where p = degree
        # m = len(knotvector) - 1
        # n = m - degree - 1
        n = num_ctrlpts - 1
        if np.abs(knotvector[n + 1] - knot) <= tol:
            return int(n)

        low = degree
        high = n + 1
        mid = int((low + high) / 2)

        while (knot < knotvector[mid]) or (knot >= knotvector[mid + 1]):
            if knot < knotvector[mid]:
                high = mid
            else:
                low = mid
            mid = int((low + high) / 2)

        return int(mid)

    @staticmethod
    @jit(nopython=True, parallel=True)
    def _find_span_array_(degree, knotvector, num_ctrlpts, knots, tol=1e-5):

        results = np.zeros(len(knots), dtype=np.int64)

        # These are knot invariant
        n = num_ctrlpts - 1
        n_knots = len(knots)

        for i in prange(n_knots):
            knot = knots[i]

            """ Algorithm A2.1 found in "The NURBS Book" by Piegl & Tiller."""
            if np.abs(knotvector[n + 1] - knot) <= tol:
                results[i] = np.int64(n)

            else:
                low = degree
                high = n + 1
                mid = int((low + high) / 2)
                while (knot < knotvector[mid]) or (knot >= knotvector[mid + 1]):
                    if knot < knotvector[mid]:
                        high = mid
                    else:
                        low = mid
                    mid = int((low + high) / 2)

                results[i] = np.int64(mid)
        return results

    @staticmethod
    @jit(nopython=True)
    def __basis_functions_ders_array__(degree=0, knotvector=(), spans=0, knots=0, order=0):
        """ Algorithm A2.3 of The NURBS Book by Piegl & Tiller."""
        # Initialize variables for easy access
        left = np.zeros(degree + 1)
        right = np.zeros(degree + 1)
        # ndu = [[None for x in range(degree+1)] for y in range(degree+1)]
        ndu = np.zeros((degree + 1, degree + 1))

        # N[0][0] = 1.0 by definition

        # ders = np.zeros(((min(degree, order) + 1), degree + 1))
        n_knots = len(knots)
        size = min(degree, order) + 1
        results = np.zeros((n_knots, size, degree + 1))

        counter = 0
        # for knot in knots:
        for counter in range(len(knots)):
            ndu[0, 0] = 1.0
            span = spans[counter]
            knot = knots[counter]

            j = int(0)
            for j in range(1, degree + 1):
                left[j] = knot - knotvector[span + 1 - j]
                right[j] = knotvector[span + j] - knot
                saved = 0.0
                r = 0
                for r in range(r, j):
                    # Lower triangle
                    ndu[j, int(r)] = right[int(r) + 1] + left[int(j) - r]
                    temp = ndu[r, j - 1] / ndu[j, r]
                    # Upper triangle
                    ndu[r][j] = saved + (right[r + 1] * temp)
                    saved = left[j - r] * temp
                ndu[j, j] = saved

            # Load the basis functions
            ders = np.zeros(((min(degree, order) + 1), degree + 1))
            for j in range(0, degree + 1):
                ders[0, j] = ndu[j, degree]

            # Start calculating derivatives
            a = np.zeros((degree + 1, 2))
            # Loop over function index
            for r in range(0, degree + 1):
                # Alternate rows in array a
                s1 = 0
                s2 = 1
                a[0, 0] = 1.0
                # Loop to compute k-th derivative
                for k in range(1, order + 1):
                    d = 0.0
                    rk = r - k
                    pk = degree - k
                    if r >= k:
                        a[s2, 0] = a[s1, 0] / ndu[pk + 1, rk]
                        d = a[s2, 0] * ndu[rk, pk]
                    if rk >= -1:
                        j1 = 1
                    else:
                        j1 = -rk
                    if (r - 1) <= pk:
                        j2 = k - 1
                    else:
                        j2 = degree - r
                    for j in range(j1, j2 + 1):
                        a[s2, j] = (a[s1, j] - a[s1, j - 1]) / ndu[pk + 1, rk + j]
                        d += (a[s2, j] * ndu[rk + j, pk])
                    if r <= pk:
                        a[s2, k] = -a[s1, k - 1] / ndu[pk + 1, r]
                        d += (a[s2, k] * ndu[r, pk])
                    ders[k, int(r)] = d

                    # Switch rows
                    j = s1
                    s1 = s2
                    s2 = j

            # Multiply through by the the correct factors
            r_fl = float(degree)
            for k in range(1, order + 1):
                for j in range(0, degree + 1):
                    ders[k, j] *= r_fl
                r_fl *= (degree - k)

            # Return the basis function derivatives list
            results[counter] = ders

        return results

    @staticmethod
    def __partial_outer_product__(n_nodes_y, n_nodes_x, n_pts, span_vs, bases_vs, span_us, bases_us, degree_v, degree_u,
                                  indices):
        # Declare results array

        results = np.zeros(n_pts * n_nodes_x * n_nodes_y, dtype=np.float64)

        # Calculate node span in x and y direction
        deg_range_u = np.array(range(0, degree_u + 1), dtype=np.int)
        deg_range_v = np.array(range(0, degree_v + 1), dtype=np.int)

        # Generate index matrix for all points
        u_ind = (span_us[:, np.newaxis] - degree_u + deg_range_u[np.newaxis, :])[:, np.newaxis, :]
        v_ind = (span_vs[:, np.newaxis] - degree_v + deg_range_v[np.newaxis, :])[:, :, np.newaxis]

        # Determine pixel indices in global frame
        index_matrix = indices[u_ind, v_ind]

        # Allow for flattening of index array
        index_matrix += (np.arange(n_pts) * n_nodes_x * n_nodes_y)[:, np.newaxis, np.newaxis]

        index_matrix_ordered = np.moveaxis(index_matrix, 0, -1)

        # Generate outer product of the derivatives vectors for a stack of vectors
        big_outer = np.einsum('ij,il->jli', bases_vs, bases_us)

        results[index_matrix_ordered] = big_outer

        return results.reshape((n_pts, -1))

    @staticmethod
    def _gen_uniform_knotvector_(degree=0, num_ctrlpts=0):
        """ Generates a uniformly-spaced knot vector using the degree and the number of control points.
        :param degree: degree of the knot vector direction
        :type degree: integer
        :param num_ctrlpts: number of control points on that direction
        :type num_ctrlpts: integer
        :return: knot vector
        :rtype: list
        """
        if degree == 0 or num_ctrlpts == 0:
            raise ValueError("Input values should be different than zero.")

        # Min and max knot vector values
        knot_min = 0.0
        knot_max = 1.0

        # Equation to use: m = n + p + 1
        # p: degree, n+1: number of ctrlpts; m+1: number of knots
        m = degree + num_ctrlpts + 1

        # Initialize return value and counter
        knotvector = []
        i = 0

        # First degree+1 knots are "knot_min"
        while i < degree + 1:
            knotvector.append(knot_min)
            i += 1

        # Calculate a uniform interval for middle knots
        num_segments = (m - (degree + 1) * 2) + 1  # number of segments in the middle
        spacing = (knot_max - knot_min) / num_segments  # spacing between the knots (uniform)
        midknot = knot_min + spacing  # first middle knot
        # Middle knots
        while i < m - (degree + 1):
            knotvector.append(midknot)
            midknot += spacing
            i += 1

        # Last degree+1 knots are "knot_max"
        while i < m:
            knotvector.append(knot_max)
            i += 1

        # Return autogenerated knot vector
        return knotvector

    @staticmethod
    @jit(nopython=True, parallel=False)
    def basis_functions_array(degree=0, knotvector=(), spans=0, knots=0):
        num_knots = len(knots)
        results = np.zeros((num_knots, degree + 1), dtype=np.float64)

        left = np.zeros(degree + 1, dtype=np.float64)
        right = np.zeros(degree + 1, dtype=np.float64)
        N = np.zeros(degree + 1, dtype=np.float64)

        for i in prange(num_knots):
            knot = knots[i]
            span = spans[i]

            """ Algorithm A2.2 of The NURBS Book by Piegl & Tiller."""

            left[:] = 0.0
            right[:] = 0.0
            N[:] = 0.0

            # N[0] = 1.0 by definition
            N[0] = 1.0

            for j in range(1, degree + 1):
                left[j] = knot - knotvector[span + 1 - j]
                right[j] = knotvector[span + j] - knot
                saved = np.float64(0.0)
                for r in range(0, j):
                    temp = N[r] / (right[r + 1] + left[j - r])
                    N[r] = saved + right[r + 1] * temp
                    saved = left[j - r] * temp
                N[j] = saved
                results[i, :] = N[:]
        return results



import random

import numpy as np
from scipy.ndimage import gaussian_filter

try:
    from noise import pnoise2
except ImportError as e:
    print(e)
    print("The package: noise, is not installed. Perlin speckle is not available")


    def pnoise2(*args, **kwargs):
        raise ImportError("The noise package is not installed")


def insert_circle(image, position, radius, allow_overlap):
    try:
        circle = np.zeros((radius * 2, radius * 2))
        xs, ys = np.meshgrid(np.arange(2 * radius), np.arange(2 * radius))
        r = ((xs - radius) ** 2. + (ys - radius) ** 2.) ** 0.5
        circle[r < radius] = 1.

        if not allow_overlap:
            non_zeros = np.count_nonzero(
                image[position[0] - radius:position[0] + radius, position[1] - radius:position[1] + radius])
            if non_zeros > 0:
                raise IndexError("Overlapping dots are not allowed")

        image[position[0] - radius:position[0] + radius, position[1] - radius:position[1] + radius] += circle
    except Exception as e:
        pass

    return image


def dots_speckle(size=(1000, 1000), n_dots=5500, dot_radius_max=40, dot_radius_min=30, blur_sigma=2,
                 allow_overlap=False):
    """ Speckle made by dots

    Returns a speckle looking like dots from a circular marker


    Example
    -------
    Let us make a speckle image with a size of 1000 x 1000 pixels, with approx 5000 dots, the smallest being 20 pixels
    and the largest being 25 pixels, all without overlap. Lets blur it to "round" the dots a little.

    The following code generates such a speckle

    >>> import muDIC as dic
    >>> speckle = dic.dots_speckle((1000,1000),n_dots=5000,dot_radius_max=25,dot_radius_min=20,blur_sigma=2,allow_overlap=False)




    Note
    ----
    The dots are not overlapping, the number of dots specified is the number of attempts to fit a dot on the image.

    Parameters
    ----------

    size=(1000,1000), n_dots=5500, dot_radius_max = 40,dot_radius_min = 30, blur_sigma=2,allow_overlap=False

    size : tuple
        The image size as a tuple of integers
    n_dots : int
        The number of dots in the image. Note that this corresponds to the number of attempts if overlap is not allowed.
    dot_radius_max: float, int
        The largest radius of a dot in the image
    dot_radius_min: float, int
        The smallest radius of a dot in the image
    blur_sigma : float
        The standard deviation of the gaussian kernel used to create gradients in the speckle image
    allow_overlap : bool
        Allow for overlapping dots
    """

    size_x, size_y = size
    img = np.zeros((size_x, size_y))

    for i in range(n_dots):
        pos_x = np.int(random.random() * size_x)
        pos_y = np.int(random.random() * size_y)

        radius = np.int(random.random() * (dot_radius_max - dot_radius_min) + dot_radius_min)

        img = insert_circle(img, (pos_x, pos_y), radius=radius, allow_overlap=allow_overlap)
    filtered = gaussian_filter(img, blur_sigma)

    filtered_normalized = normalize_array_to_unity(filtered)

    return filtered_normalized * -1. + 1.


def harmonic_speckle(size=(1000, 1000), n_peaks_x=20):
    """ Speckle made by harmonic functions

    Returns a speckle looking like a bilateral wave pattern


    Example
    -------
    Let us make a speckle image with a size of 1000 x 1000 pixels, with 20 peaks along the first axis

    The following code generates such a speckle

    >>> import muDIC as dic
    >>> speckle = dic.harmonic_speckle((1000,1000),n_peaks_x=20)


    Parameters
    ----------

    size=(1000,1000), n_dots=5500, dot_radius_max = 40,dot_radius_min = 30, blur_sigma=2,allow_overlap=False

    size : tuple
        The image size as a tuple of integers
    n_peaks_x : int
        The number of peaks along the first axis of the image
    """
    size_x, size_y = size
    xs, ys = np.meshgrid(np.arange(size_x), np.arange(size_y))

    freq = np.pi * 2. * np.float(n_peaks_x) / size_x
    x_harm = np.sin(xs * freq)
    y_harm = np.sin(ys * freq)

    field = x_harm * y_harm

    return normalize_array_to_unity(field)


def normalize_array_to_unity(array):
    return (array - np.min(array)) / (np.max(array) - np.min(array))


def smooth_step(array, c):
    return 0.5 * (1. + np.tanh((array) / c ** 2.))


def perlin_noise_speckle(shape, multiplier=64., octaves=1):
    """ Perlin noise based speckle

    Returns a speckle made by using Perlin noise provided by the Noise package by Casey Duncan (caseman).


    Example
    -------
    Let us make an example perlin noise spackle with a size of 1000x1000 pixels using the defaults.

    The following code generates such a speckle

    >>> import muDIC as dic
    >>> speckle = dic.perlin_noise_speckle((1000,1000))

    Notes
    -------
    The speckle generator uses the "pnoise2" function of the noise library, so you can look at the docs for that library
    for further documentation.

    Parameters
    ----------

    shape : tuple
        The image size as a tuple of integers
    multiplier : float
        The frequency multiplier
    octaves: float, int
        The number of octaves used
    """

    freq = multiplier * octaves
    img = np.zeros(shape)
    n, m = shape

    for y in range(n):
        for x in range(m):
            img[x, y] = float(pnoise2(x / freq, y / freq, octaves) * (float(n) - 1.) + float(n))

    img = normalize_array_to_unity(img) * 2.
    img = smooth_step(img.astype(np.float), c=0.7)
    img = normalize_array_to_unity(img)

    return img


def rosta_speckle(size, dot_size=4, density=0.32, smoothness=2.0):
    """ Rosta speckle

     Returns a speckle made by the "Rosta" algorithm
     This algorithm is very pragmatic and makes a speckle looking like a "real" spray speckle without
     being based on any real physics.



     Example
     -------
     Let us make a speckle image with a size of 1000 x 1000 pixels, using the default values.

     The following code generates such a speckle

     >>> import muDIC as dic
     >>> speckle = rosta_speckle((1000,1000), dot_size=4, density=0.32, smoothness=2.0, layers=1)

    If you want a denser speckle, you can increase the number of layers like this:

     >>> speckle = rosta_speckle((1000,1000), dot_size=4, density=0.32, smoothness=2.0, layers=1)


     Parameters
     ----------

     size : tuple
         The image size as a tuple of integers
     dot_size : int
         The size of the dots [0,1]
     density: float, int
        How packeg with dots the speckle should be
     smoothness: float, int
        The degree of smoothing applied to the binary speckle
     """
    merge_sigma = dot_size * size[0] / 1000.
    blur_sigma = smoothness * size[0] / 1000.

    noise = np.random.randn(*size)
    noise_blurred = gaussian_filter(noise, sigma=merge_sigma)
    noise_blurred = normalize_array_to_unity(noise_blurred)

    sorted_gray_scales = np.sort(noise_blurred.flatten())

    clip_index = int(density * np.size(sorted_gray_scales))
    clipval = sorted_gray_scales[clip_index]

    clipped = np.zeros_like(noise_blurred)
    clipped[noise_blurred > clipval] = 1.0

    speckle = gaussian_filter(clipped, sigma=blur_sigma) * -1. + 1.
    return speckle



from abc import abstractmethod, ABCMeta


class FieldInterpolator(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def Nn(self, e, n):
        raise NotImplementedError

    @abstractmethod
    def dxNn(self, e, n):
        raise NotImplementedError

    @abstractmethod
    def dyNn(self, e, n):
        raise NotImplementedError



import muDIC as dic
import logging

# Set the amount of info printed to terminal during analysis
logging.basicConfig(format='%(name)s:%(levelname)s:%(message)s', level=logging.INFO)

# Path to folder containing images
path = r'./example_data/'

# Generate image instance containing all images found in the folder
images = dic.IO.image_stack_from_folder(path, file_type='.tif')
images.set_filter(dic.filtering.lowpass_gaussian, sigma=1.)


# Generate mesh
mesher = dic.Mesher(deg_e=3, deg_n=3)
mesh = mesher.mesh(images,Xc1=200,Xc2=1050,Yc1=200,Yc2=650,n_ely=8,n_elx=8, GUI=False)


# Instantiate settings object and set some settings manually
settings = dic.solver.correlate.DICInput(mesh, images)
settings.max_nr_im = 40
settings.ref_update = [15]
settings.maxit = 20
# If you want to access the residual fields after the analysis, this should be set to True
settings.store_internals = False

# Instantiate job object
job = dic.solver.DICAnalysis(settings)

# Running DIC analysis
dic_results = job.run()

# Calculate field values
fields = dic.post.viz.Fields(dic_results)

# Show a field
viz = dic.Visualizer(fields,images=images)
viz.show(field="true strain", component = (1,1), frame = 39)




from functools import partial

import numpy as np


def gaussian_noise_model(size, sigma):
    return np.random.normal(0.0, sigma, size)


def noise_injector_additive(image, noise_generator):
    mean = np.average(image)

    image_noisy = image + mean * noise_generator(image.shape)
    return image_noisy


def noise_injector(noise_model="gaussian", sigma=0.0):
    """Noise injector

    The noise injector returns a function which when passed an image, adds noise to the image.

    The supported noise models are:
        * Gaussian additive noise

    Example
    -------
    Let us make an noise injector which adds gaussian noise with a standard deviation of 1.

    >>> from muDIC import vlab
    >>> import numpy as np
    >>> img = np.ones((100,100))
    >>> noise_injector = vlab.noise_injector("gaussian", sigma=.1)
    >>> noisy_image = noise_injector(img)


    Note
    ----
    More noise models should be added later

    Parameters
    ----------
    noise_model : string
        The noise model to be used. Eg. "gaussian"
    sigma : float
        The standard deviation of the noise distribution
    """
    if noise_model == "gaussian":
        noise_generator = partial(gaussian_noise_model, sigma=sigma)

    return partial(noise_injector_additive, noise_generator=noise_generator)



import logging
from functools import partial

import numpy as np

from ..utils import find_borders, find_inconsistent

np.seterr(invalid='raise')


class Reference(object):
    def __init__(self, Nref, I0, K, B, n_pixels, e, n, image_id=None):
        """
        Reference container

        Contains all neccessary data related to the reference image.

        Parameters
        ----------
        Nref : ndarray
            The shape function of the element evaluated at all image coordinates
        I0 : ndarray
            The pixel values for all pixels covered by the mesh
        K : ndarray.
            Matrix used during the correlation process
        B : ndarray
            Matrix used during the correlation process
        n_pixels : int
            The number of pixels covered by the mesh
        e : ndarray.
            The e coodinate to the pixels covered by the mesh
        n : ndarray
            The n coodinate to the pixels covered by the mesh
        image_id : int, optional
            The id for the image used to make the reference

        Returns
        -------
        Refernce object


        Notes
        -----
        This constructor does not do any type of typechecking
        """
        self.Nref_stack = Nref
        self.I0_stack = I0
        self.K = K
        self.B_stack = B
        self.n_pixels = n_pixels
        self.e = e
        self.n = n
        self.image_id = image_id


def find_active_pixels(e_coord, n_coord, e_coord_inc, n_coord_inc, min_inc=1e-6):
    """
    Finds active pixels

    Returns a boolean array identifying which pixels have increments larger than the tolerance
    and within the bounds [0,1] of the frame.


    Parameters
    ----------
    e_coord : ndarray
        The e coordinates of the pixel
    n_coord : ndarray
        The n coordinates of the pixel
    e_coord_inc : ndarray.
        The e coordinate increment of the pixel
    n_coord_inc : ndarray
        The n coordinate increment of the pixel
    min_inc : float
        The increment size which defines convergence

    Returns
    -------
    active_pixels : ndarry
        The active pixels given as a 1d boolean array


    Notes
    -----
    """
    e_coord_inbound = np.logical_and(e_coord > 0., e_coord < 1.)
    n_coord_inbound = np.logical_and(n_coord > 0., n_coord < 1.)
    not_converged = np.logical_or(np.abs(e_coord_inc) > min_inc, np.abs(n_coord_inc) > min_inc)

    return np.logical_and(np.logical_and(e_coord_inbound, n_coord_inbound), not_converged)


def clip_args(func, arg1, arg2, bounds=(0., 1.)):
    """
    Clip the arguments to bounds

    Return the results of the function where the clipped arguments have been used.
    Arguments below the lower bound are set to the lower bound and the arguments
    above the upper bound are set to the upper bound.

    Parameters
    ----------
    func : func(arg1,arg2)
        The function which the clipped arguments are passed to
    arg1 : ndarray
        1D array with floats.
    arg2 : ndarray.
        1D array with floats.
    bounds : tuple, optional
        The bounds that the arguments are limited to.

    Returns
    -------
    clipped_func : func(arg1_clipped,arg2_clipped)
        The results of the function where the clipped agruments have been applied.


    Notes
    -----
    This function does not do any type of typechecking
    """
    upper_bound = bounds[1]
    lower_bound = bounds[0]

    arg1_inbound = arg1.copy()
    arg2_inbound = arg2.copy()

    arg1_inbound[arg1 < lower_bound] = lower_bound
    arg1_inbound[arg1 > upper_bound] = upper_bound

    arg2_inbound[arg2 < lower_bound] = lower_bound
    arg2_inbound[arg2 > upper_bound] = upper_bound

    return func(arg1_inbound, arg2_inbound)


def identify_pixels_within_frame(xnod, ynod, elm, over_sampling=1.1):
    """
        Identify pixels covered by an element frame.

        Returns the coordinates of the covered pixels in the image frame and an estimate of the coordinates
        in the element frame.
        This is done by evaluating the element shape functions on a denser grid than the
        image grid, rounds the indices to nearest integer and removes duplicates.
        The element cordinates to the corresponding pixels is then obtained from the same mask.

        Parameters
        ----------
        xnod : ndarray
            1D array with floats.
            The x coordinates of the control points.
        ynod : ndarray
            1D array with floats.
            The y coordinates of the control points.
        elm : interpolator object.
            The interpolator object provides the shape functions used to calculate the coordinates within the element.
        over_sampling : float, optional
            The degree of oversampling used to find the pixels.

        Returns
        -------
        pixel_x : ndarray
            The x-coordinates of the pixels covered by the element
        pixel_y : ndarray
            The y-coordinates of the pixels covered by the element
        pixel_es : ndarray
            The elemental e-coordinates of the pixels covered by the element
        pixel_ns : ndarray
            The elemental n-coordinates of the pixels covered by the element


        Notes
        -----
        There is no guarantee that all pixels are found, so when in doubt, increase the over_sampling factor.

        """
    x_min, x_max = find_borders(xnod)
    y_min, y_max = find_borders(ynod)

    # Calculate coordinates (e,n) covered by the element on a fine grid
    n_search_pixels = np.int(over_sampling * max((x_max - x_min), y_max - y_min))
    es, ns = np.meshgrid(np.linspace(0., 1., n_search_pixels), np.linspace(0., 1., n_search_pixels))
    es = es.flatten()
    ns = ns.flatten()

    pixel_xs = np.dot(elm.Nn(es, ns), xnod)
    pixel_ys = np.dot(elm.Nn(es, ns), ynod)

    pixel_xs_closest = np.around(pixel_xs).astype(np.int)
    pixel_ys_closest = np.around(pixel_ys).astype(np.int)

    xs_ys = np.stack([pixel_xs_closest, pixel_ys_closest], axis=0)
    xs_ys_unique, unique_inds = np.unique(xs_ys, return_index=True, axis=1)

    pixel_x = xs_ys_unique[0, :].astype(np.float64)
    pixel_y = xs_ys_unique[1, :].astype(np.float64)

    pixel_es = es[unique_inds].astype(np.float64)
    pixel_ns = ns[unique_inds].astype(np.float64)

    return pixel_x, pixel_y, pixel_es, pixel_ns


def find_covered_pixel_blocks(node_x, node_y, elm, max_iter=200, block_size=1e7, tol=1.e-6):
    """
    Find element coordinates to all pixels covered by the element.

    Returns the coordinates of the covered pixels in the image coordinates and in the element coordinates.
    This is done by first identifiying the pixels within the frame and then finding the corresponding
    element coordinates by using a modified Newton-Raphson scheme. For reduced memory usage, the
    image covered by the element is subdivided into blocks.

    Parameters
    ----------
    node_x : ndarray
        1D array with floats.
        The x coordinates of the control points.
    node_y : ndarray
        1D array with floats.
        The y coordinates of the control points.
    elm : interpolator object.
        The interpolator object provides the shape functions used to calculate the coordinates within the element.
    max_iter : int, optional
        The maximum allowed number of iterations
    block_size :int, optional
        The maximum number of elements in each block
        The number of elements are N-pixels X N-Control points

    tol : float, optional
        The convergence criteria

    Returns
    -------
    pixel_x : ndarray
        The x-coordinates of the pixels covered by the element
    pixel_y : ndarray
        The y-coordinates of the pixels covered by the element
    pixel_es : ndarray
        The elemental e-coordinates of the pixels covered by the element
    pixel_ns : ndarray
        The elemental n-coordinates of the pixels covered by the element


    Notes
    -----

    """

    logger = logging.getLogger(__name__)

    # e and n are element coordinates
    found_e = []
    founc_n = []
    # x and y are the corresponding image coordinates
    found_x = []
    found_y = []

    # These are just estimates
    pix_Xs, pix_Ys, pix_es, pix_ns = identify_pixels_within_frame(node_x, node_y, elm)

    # Split into blocks
    n_pix_in_block = block_size / np.float(len(node_x))
    num_blocks = np.ceil(len(pix_es) / n_pix_in_block).astype(np.int)
    logger.info("Splitting in %s blocks:" % str(num_blocks))

    pix_e_blocks = np.array_split(pix_es, num_blocks)
    pix_n_blocks = np.array_split(pix_ns, num_blocks)

    pix_X_blocks = np.array_split(pix_Xs, num_blocks)
    pix_Y_blocks = np.array_split(pix_Ys, num_blocks)

    for block_id in range(num_blocks):
        e_coord = pix_e_blocks[block_id]
        n_coord = pix_n_blocks[block_id]

        X_coord = pix_X_blocks[block_id]
        Y_coord = pix_Y_blocks[block_id]

        # Empty increment vectors
        n_coord_inc = np.zeros_like(n_coord)
        e_coord_inc = np.zeros_like(e_coord)

        # Pre-calculate the gradients. This results in a modified Newton scheme
        dxNn = clip_args(elm.dxNn, e_coord, n_coord)
        dyNn = clip_args(elm.dyNn, e_coord, n_coord)

        for i in range(max_iter):

            Nn = clip_args(elm.Nn, e_coord, n_coord)

            n_coord_inc[:] = (Y_coord - np.dot(Nn, node_y) - np.dot(dxNn, node_y) * (
                    X_coord - np.dot(Nn, node_x)) / (np.dot(dxNn, node_x))) / (
                                     np.dot(dyNn, node_y) - np.dot(dxNn, node_y) * np.dot(
                                 dyNn, node_x) / np.dot(dxNn, node_x))

            e_coord_inc[:] = (X_coord - np.dot(Nn, node_x) - np.dot(dxNn, node_x) *
                              n_coord_inc) / np.dot(dxNn, node_x)

            e_coord[:] += e_coord_inc
            n_coord[:] += n_coord_inc

            active_pixels = find_active_pixels(e_coord, n_coord, e_coord_inc, n_coord_inc, tol)

            if not np.any(active_pixels):
                logger.info('Pixel coordinates found in %i iterations', i)

                epE_block, nyE_block, Xe_block, Ye_block = map(
                    partial(np.delete, obj=find_inconsistent(e_coord, n_coord)),
                    [e_coord, n_coord, X_coord, Y_coord])

                found_e.append(epE_block)
                founc_n.append(nyE_block)
                found_x.append(Xe_block.astype(np.int))
                found_y.append(Ye_block.astype(np.int))

                break
            if (i + 1) == max_iter:
                raise RuntimeError("Did not converge in %i iterations" % max_iter)

    return found_e, founc_n, found_x, found_y


def generate_reference(nodal_position, mesh, image, settings, image_id=None):
    """
    Generates a Reference object

    The Reference object contains all internals that will be used during the correlation procedure.


   Parameters
   ----------
   nodal_position : ndarray
       2D array with floats.
       The coordinates of the control points.
   mesh : Mesh object
       Mesh definitions
   image : ndarray
       image as an 2d array
   image_id : int, optional
       The image id, stored for further reference


   Returns
   -------
   reference : Reference
       The Reference object

   Notes
   -----
   The current implementation is slow but not very memory intensive.

   Theory
   -----


   """
    logger = logging.getLogger()
    elm = mesh.element_def

    node_xs = nodal_position[0]
    node_ys = nodal_position[1]

    img_grad = np.gradient(image)

    try:
        pixel_e_blocks, pixel_n_blocks, pixel_x_blocks, pixel_y_blocks = find_covered_pixel_blocks(node_xs,
                                                                                                   node_ys,
                                                                                                   elm,
                                                                                                   block_size=settings.block_size)

        num_blocks = len(pixel_e_blocks)
        num_pixels = np.sum([block.size for block in pixel_e_blocks])

        K = np.zeros((2 * mesh.n_nodes, num_pixels), dtype=settings.precision)
        A = np.zeros((mesh.n_nodes * 2, mesh.n_nodes * 2), dtype=settings.precision)

        img_covered = image[np.concatenate(pixel_y_blocks), np.concatenate(pixel_x_blocks)]

        # Calculate A = B^T * B
        for block_id in range(num_blocks):
            block_len = pixel_e_blocks[block_id].shape[0]
            B = np.zeros((block_len, 2 * mesh.n_nodes), dtype=settings.precision)

            # Weight the image gradients with the value of the shape functions
            B[:, :elm.n_nodes] = (
                    img_grad[1][pixel_y_blocks[block_id], pixel_x_blocks[block_id]][:, np.newaxis] * elm.Nn(
                pixel_e_blocks[block_id], pixel_n_blocks[block_id]))

            B[:, elm.n_nodes:] = (
                    img_grad[0][pixel_y_blocks[block_id], pixel_x_blocks[block_id]][:, np.newaxis] * elm.Nn(
                pixel_e_blocks[block_id], pixel_n_blocks[block_id]))
            A += np.dot(B.transpose(), B)

        pixel_ind = 0
        pixel_ind_last = 0

        # Determine K
        for block_id in range(num_blocks):
            block_len = pixel_e_blocks[block_id].shape[0]

            B = np.zeros((2 * mesh.n_nodes, block_len), dtype=settings.precision)

            pixel_ind += block_len

            # Weight the image gradients with the value of the shape functions
            # TODO: This operation is duplicate
            B[:elm.n_nodes, :] = (
                    img_grad[1][pixel_y_blocks[block_id], pixel_x_blocks[block_id]][:, np.newaxis] * elm.Nn(
                pixel_e_blocks[block_id], pixel_n_blocks[block_id])).transpose()

            B[elm.n_nodes:, :] = (
                    img_grad[0][pixel_y_blocks[block_id], pixel_x_blocks[block_id]][:, np.newaxis] * elm.Nn(
                pixel_e_blocks[block_id], pixel_n_blocks[block_id])).transpose()

            K_block = np.linalg.solve(A, B)

            K[:, pixel_ind_last:pixel_ind] = K_block

            pixel_ind_last = pixel_ind

        # Remove for reduced memory usage
        del B, K_block

        Nn = elm.Nn(np.concatenate(pixel_e_blocks), np.concatenate(pixel_n_blocks)).transpose()
        pixel_es = np.concatenate(pixel_e_blocks)
        pixel_ns = np.concatenate(pixel_n_blocks)


    except Exception as e:
        logger.exception(e)
        raise RuntimeError('Failed to generate reference')

    return Reference(Nn, img_covered, K, None, num_pixels, pixel_es, pixel_ns,
                     image_id=image_id)



from . import deformation_fields
from .downsampler import Downsampler
from .image_deformer import imageDeformer_from_defGrad, imageDeformer_from_uFunc
from .noise import noise_injector
from .speckle import dots_speckle, rosta_speckle, harmonic_speckle, perlin_noise_speckle
from .syntheticImageGenerator import SyntheticImageGenerator



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="muDIC",
    version="0.0.7",
    author="PolymerGuy",
    author_email="sindre.n.olufsen@ntnu.no",
    description="A digital image correlation toolkit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/PolymerGuy/muDIC",
    packages=setuptools.find_packages(),
    install_requires=[
        'numba',
        'scipy >= 1.2.1',
        'matplotlib',
        'numpy',
        'Pillow',
        'dill',
        'nose',
        'scikit-image',
        'muDIC'
      ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import logging

import matplotlib.pyplot as plt
import numpy as np


class Fields(object):

    def __init__(self, dic_results, seed=21):
        """
        Fields calculates field variables from the DIC-results.
        The implementation is lazy, hence getter methods have to be used.

        NOTE
        ----
        The fields are formatted as follows:

        * Vectors: [elm_id,component_i,element_e_coord,element_n_coord,frame_id]
        * matrices: [elm_id,component_i,component_j,element_e_coord,element_n_coord,frame_id]


        Parameters
        ----------
        dic_results :
            The results from the DIC analysis
        seed : Integer
            The number of grid points which will be evaluated in each direction

        Returns
        -------
        A Fields object
        """

        self.logger = logging.getLogger()

        # The type is implicitly checked by using the interface
        self.__res__ = dic_results
        self.__settings__ = dic_results.settings

        self.__ee__, self.__nn__ = self.__generate_grid__(seed)

        self.__F__, self.__coords__ = self._deformation_gradient_(self.__res__.xnodesT, self.__res__.ynodesT,
                                                                  self.__settings__.mesh,
                                                                  self.__settings__.mesh.element_def, self.__nn__,
                                                                  self.__ee__)

    def __generate_grid__(self, seed):
        self.__inc__ = 1. / (float(seed) - 1.)

        return np.meshgrid(np.arange(0., 1. + self.__inc__, self.__inc__),
                           np.arange(0., 1. + self.__inc__, self.__inc__))

    @staticmethod
    def _deformation_gradient_(xnodesT, ynodesT, msh, elm, e, n):
        """
        Calculate the deformation gradient from the control point positions
        and the element definitions.

        See the paper for the procedure.

        Parameters
        ----------
        xnodesT : ndarray
            Node position in the x direction
        ynodesT : ndarray
            Node position in the y direction
        msh : Mesh
            A Mesh object
        elm : Element
            A Element object containing the element definitions
        e : ndarray
            The e coordinates of the element
        n : ndarray
            The n coordinates of the element
        """

        # Post Processing
        nEl = 1
        ne = np.shape(e)[0]
        nn = np.shape(e)[1]

        # Evaluate shape function gradients on grid within element
        Nn = elm.Nn(e.flatten(), n.flatten())
        dfde = elm.dxNn(e.flatten(), n.flatten())
        dfdn = elm.dyNn(e.flatten(), n.flatten())

        Fstack = []
        coord_stack = []

        for el in range(nEl):
            x_crd = np.einsum('ij,jn -> in', Nn, xnodesT)

            y_crd = np.einsum('ij,jn -> in', Nn, ynodesT)

            dxde = np.einsum('ij,jn -> in', dfde, xnodesT)

            dxdn = np.einsum('ij,jn -> in', dfdn, xnodesT)

            dyde = np.einsum('ij,jn -> in', dfde, ynodesT)

            dydn = np.einsum('ij,jn -> in', dfdn, ynodesT)

            c_confs = np.array([[dxde, dxdn], [dyde, dydn]])
            r_conf_inv = np.linalg.inv(np.rollaxis(c_confs[:, :, :, 0], 2, 0))

            Fs = np.einsum('ijpn,pjk->ikpn', c_confs, r_conf_inv)

            Fs = Fs.reshape((2, 2, ne, nn, -1))

            x_crd = x_crd.reshape((ne, nn, -1))
            y_crd = y_crd.reshape((ne, nn, -1))

            Fstack.append(Fs)

            coord_stack.append(np.array([x_crd, y_crd]))

        # Returns F(nElms, i, j, ide, idn , frame), coords(nElms, i, ide, idn , frame)

        return np.array(Fstack), np.array(coord_stack)

    @staticmethod
    def _green_deformation_(F):
        """
        Calculate Green deformation tensor from deformation as G = F^T * F
        :param F:
        :return:
        """
        E11 = F[:, 0, 0, :, :, :] ** 2. + F[:, 0, 1, :, :, :] ** 2.

        E12 = F[:, 0, 0, :, :, :] * F[:, 1, 0, :, :, :] + F[:, 0, 1, :, :, :] * F[:, 1, 1, :, :, :]

        E22 = F[:, 1, 0, :, :, :] ** 2. + F[:, 1, 1, :, :, :] ** 2.

        E = np.array([[E11, E12], [E12, E22]])

        E[E == np.nan] = 0.

        return np.moveaxis(E, 2, 0)

    @staticmethod
    def _green_strain_(F):
        """
        Calculate Green strain tensor from F as G = 0.5*(F^T * F -I)
        :param F: Deformation gradient tensor F_ij on the form [nEl,i,j,...]
        :return: Green Lagrange strain tensor E_ij on the form [nEl,i,j,...]
        """
        E11 = 0.5 * (F[:, 0, 0, :, :, :] ** 2. + F[:, 0, 1, :, :, :] ** 2. - 1.)

        E12 = 0.5 * (F[:, 0, 0, :, :, :] * F[:, 1, 0, :, :, :] + F[:, 0, 1, :, :, :] * F[:, 1, 1, :, :, :])

        E22 = 0.5 * (F[:, 1, 0, :, :, :] ** 2. + F[:, 1, 1, :, :, :] ** 2. - 1.)

        E = np.array([[E11, E12], [E12, E22]])

        E[E == np.nan] = 0.

        return np.moveaxis(E, 2, 0)

    @staticmethod
    def _principal_strain_(G):
        E11 = G[:, 0, 0]
        E12 = G[:, 0, 1]
        E21 = G[:, 1, 0]
        E22 = G[:, 1, 1]

        E_temp = np.moveaxis(G, 1, -1)
        E = np.moveaxis(E_temp, 1, -1)

        eigvals, eigvecs = np.linalg.eig(E)

        # print(np.shape(eigvals))
        # print(np.shape(eigvecs))

        ld1 = np.sqrt(eigvals[:, :, :, :, 0])
        ld2 = np.sqrt(eigvals[:, :, :, :, 1])

        ev1 = eigvecs[:, :, :, :, 0, 0]
        ev2 = eigvecs[:, :, :, :, 0, 1]

        # print(np.shape(eigvals))
        # print(np.shape(eigvecs))
        # print(np.shape(ld1))
        # print(np.shape(ev1))

        ld = np.moveaxis(np.array([ld1, ld2]), 0, 1)
        ev = np.moveaxis(np.array([ev1, ev2]), 0, 1)
        print(np.shape(ld1))
        print(np.shape(ev1))

        return ld, ev

    @staticmethod
    def _engineering_strain_(E):
        """
        Calculate engineering strain from Green Lagrange strain tensor E_ij as:
        eps_ii = sqrt(1+E_ii)-1 and
        gamma_ij = 2E_ij/sqrt((1+E_ii)*(1+E_jj))
        :param E: Green Lagrange strain tensor E_ij on the form [nEl,i,j,...]
        :return: Engineering strain tensor eps_ij on the form [nEl,i,j,...]
        """
        eps_xx = np.sqrt(1. + 2. * E[:, 0, 0, :]) - 1.
        eps_yy = np.sqrt(1. + 2. * E[:, 1, 1, :]) - 1.
        eps_xy = 0.5 * np.arcsin(2. * E[:, 0, 1, :] / np.sqrt((1. + 2. * E[:, 0, 0, :]) * (1. + 2. * E[:, 1, 1, :])))

        eps = np.array([[eps_xx, eps_xy], [eps_xy, eps_yy]])

        return np.moveaxis(eps, 2, 0)

    @staticmethod
    def _true_strain_(eps):
        """
        Calculate true strain tensor teps_ij from engineering strain tensor eps_ij as:
        teps_ij = log(eps_ij+1)
        :param eps: Engineering strain tensor eps_ij on the form [nEl,i,j,...]
        :return: True strain tensor teps_ij on the form [nEl,i,j,...]
        """
        return np.log(eps + 1.)

    def true_strain(self):
        E = self._green_strain_(self.__F__)
        engineering_strains = self._engineering_strain_(E)
        return self._true_strain_(engineering_strains)

    def eng_strain(self):
        E = self._green_strain_(self.__F__)
        return self._engineering_strain_(E)

    def F(self):
        return self.__F__

    def green_strain(self):
        return self._green_strain_(self.__F__)

    def coords(self):
        return self.__coords__

    def disp(self):
        return self.__coords__[:, :, :, :, :] - self.__coords__[:, :, :, :, 0, np.newaxis]

    def residual(self, frame_id):
        if self.__settings__.store_internals ==False:
            raise ValueError("The analysis has to be run with store_internals=True")
        ref_id = ind_closest_below(frame_id, [ref.image_id for ref in self.__res__.reference])
        ref = self.__res__.reference[ref_id]

        cross_correlation_product = cross_correlation_products(self.__res__.Ic_stack[frame_id], ref.I0_stack)
        self.logger.info("Cross correlation product is %f" % cross_correlation_product)

        return np.abs(self.__res__.Ic_stack[frame_id] - ref.I0_stack)

    def elm_coords(self, frame_id):
        ref_id = ind_closest_below(frame_id, [ref.image_id for ref in self.__res__.reference])
        ref = self.__res__.reference[ref_id]
        return ref.e, ref.n


class Visualizer(object):
    def __init__(self, fields, images=False):
        """
        Visualizer for field variables.

        Parameters
        ----------
        fields : Fields object
            The Fields object contains all the variables that can be plotted.
        images : ImageStack object
            The stack of images corresponding to Fields

        Returns
        -------
        A Visualizer Object
        """
        if isinstance(fields, Fields):
            self.fields = fields
        else:
            raise ValueError("Only instances of Fields are accepted")

        self.images = images
        self.logger = logging.getLogger()

    def show(self, field, component=(0, 0), frame=0):
        """
        Show the field variable

        Parameters
        ----------
        field : string
            The name of the field to be shown. Valid inputs are:
                "true strain"
                "eng strain"
                "disp"
                "green strain"
                "residual"

        component : tuple with length 2
            The components of the fields. Ex. (0,1).
            In the case of vector fields, only the first index is used.
        frame : Integer
            The frame number of the field

        """

        keyword = field.replace(" ", "").lower()

        if keyword == "truestrain":
            fvar = self.fields.true_strain()[0, component[0], component[1], :, :, frame]
            xs, ys = self.fields.coords()[0, 0, :, :, frame], self.fields.coords()[0, 1, :, :, frame]

        elif keyword == "engstrain":
            fvar = self.fields.eng_strain()[0, component[0], component[1], :, :, frame]
            xs, ys = self.fields.coords()[0, 0, :, :, frame], self.fields.coords()[0, 1, :, :, frame]

        elif keyword in ("displacement", "disp", "u"):
            fvar = self.fields.disp()[0, component[0], :, :, frame]
            xs, ys = self.fields.coords()[0, 0, :, :, frame], self.fields.coords()[0, 1, :, :, frame]

        elif keyword == "greenstrain":
            fvar = self.fields.green_strain()[0, component[0], component[1], :, :, frame]
            xs, ys = self.fields.coords()[0, 0, :, :, frame], self.fields.coords()[0, 1, :, :, frame]

        elif keyword == "residual":
            fvar = self.fields.residual(frame)
            xs, ys = self.fields.elm_coords(frame)

        else:
            self.logger.info("No valid field name was specified")
            return

        if np.ndim(fvar) == 2:
            if self.images:
                plt.imshow(self.images[frame], cmap=plt.cm.gray)
            plt.contourf(xs, ys, fvar, 50, alpha=0.8)
        else:
            plt.tricontourf(xs, ys, fvar, 50, alpha=0.8)

        plt.colorbar()
        plt.show()


def ind_closest_below(value, list):
    ind = 0
    for i, num in enumerate(list):
        if num < value:
            ind = i

    return ind


def cross_correlation_products(field_a, field_b):
    return np.sum(field_a * field_b) / (
            (np.sum(np.square(field_a)) ** 0.5) * (
            np.sum(np.square(field_b)) ** 0.5))



import numpy as np

from . import image_deformer as cordmap
from .downsampler import Downsampler


class SyntheticImageGenerator(object):
    def __init__(self, speckle_image, image_deformer, downsampler, noise_injector, n=10):
        """Virtual experiment

        This class sets up a simple pipeline for generating a set of synthetically generated
        and deformed speckle images.

        The pipeline is as follows:
        A speckle image is deformed by the image_deformer
        The deformed image is then downsampled
        The downsampled image is then added noise

        For a larger deformation, the image_deformer deforms the image more and the same pipeline is used.

        Example
        -------
        Let us generate a synthetic image generator which produced 500x500 pixel images with the following approach

        * Generate a synthetic speckle using the "Rosta" algorithm
        * Deform the base image by a given deformation gradient
        * Down-sample the image by a factor of 4, with a pixel fill factor of 0.8 and a pixel offset standard deviation
            of 0.1 pixel
        * Add gaussian additive noise to the image

        The neccessary code to do this is as follows:

            >>> n = 10
            >>> image_shape = (2000, 2000)
            >>> # speckle_image = vlab.speckle.make_perlin_noise_speckle(image_shape,octaves=3)
            >>> speckle_image = vlab.rosta_speckle(image_shape, dot_size=4, density=0.32, smoothness=2.0, layers=4)
            >>>
            >>> F = np.array([[1.1, .0], [0., 1.0]], dtype=np.float64)
            >>> image_deformer = vlab.imageDeformer_from_defGrad(F)
            >>>
            >>> downsampler = vlab.Downsampler(image_shape=image_shape, factor=4, fill=0.8, pixel_offset_stddev=0.1)
            >>>
            >>> noise_injector = vlab.noise_injector("gaussian", sigma=.01)
            >>>
            >>> image_generator = vlab.SyntheticImageGenerator(speckle_image=speckle_image, image_deformer=image_deformer,
            >>>                                          downsampler=downsampler, noise_injector=noise_injector, n=n)
            >>> image_stack = dic.ImageStack(image_generator)



        Note
        ----
        All arguments have to be provided and no default behavior is defined

        Parameters
        ----------
        Speckle_image : ndarray
            The speckle image.
        image_deformer : instance of ImageDeformer
            ImageDeformer deforms an image according to the coordinate mapper given to it upon instantiation
        downsampler: instance of Downsampler
            Downsampler downsamples an image according to the settings given to it upon instantiation
        noise_injector: noise injection function
            noise_injector is a function which takes an image and returns an image with noise
        n : int
            Number of frames to be deformed
        """

        self.n_images = n

        if isinstance(speckle_image, np.ndarray):
            if speckle_image.ndim != 2:
                raise ValueError("Only 2D arrays are accepted")
            self.speckle_image = speckle_image

        if isinstance(image_deformer, cordmap.ImageDeformer):
            self._image_deformer_ = image_deformer
        else:
            raise ValueError("Only instances of CoordinateMapper are accepted")

        # If a speckle image is provided, generate a matching down sampler
        if isinstance(downsampler, Downsampler):
            self.down_sampler = downsampler
        else:
            raise ValueError("Only instances of Downsampler are accepted")

        if noise_injector is None:
            raise ValueError("Noise_injector must be specified")



        self.noise_injector = noise_injector

        self._deformed_images_ = self._image_deformer_(self.speckle_image, steps=self.n_images)

        self._deformed_noisy_images_ = [self.noise_injector(img) for img in self._deformed_images_]

        self.phantoms = [self.down_sampler(img) for img in self._deformed_noisy_images_]

    def deformation_field(self):
        # return self._generate_F_field_(self.F, self._weight_field_)
        # return np.linalg.inv(self._F_field_)
        return self._image_deformer_.coodinate_mapper.def_grad

    def __len__(self):
        return len(self.phantoms)

    def __call__(self, img_ind):
        # TODO: Make this thing lazy
        return self.phantoms[img_ind]
